"""Reads and writes XML for Axiell EMu"""

from .containers import (
    EMuColumn,
    EMuConfig,
    EMuEncoder,
    EMuGrid,
    EMuRow,
    EMuRecord,
    EMuSchema,
)
from .io import EMuReader, write_csv, write_group, write_import, write_xml
from .types import EMuDate, EMuFloat, EMuLatitude, EMuLongitude, EMuTime, EMuType
from .utils import (
    flatten,
    get_mod,
    has_mod,
    is_nesttab,
    is_nesttab_inner,
    is_ref,
    is_ref_tab,
    is_tab,
    strip_mod,
    strip_tab,
)

__version__ = "0.1b7"
__author__ = "Adam Mansur"
__credits__ = "Smithsonian National Museum of Natural History"



"""Functions to assess whether EMu field names are tables, references, etc."""

import re
from functools import lru_cache

#: tuple : suffixes that designate tables in EMu
TAB_SUFFIXES = ("0", "_nesttab", "_nesttab_inner", "_tab")

#: tuple : suffixes that designate references in EMu
REF_SUFFIXES = ("Ref", "Ref_nesttab", "Ref_nesttab_inner", "Ref_tab")

#: tuple : suffixes that designate nested tables in EMu
NESTTAB_SUFFIXES = ("_nesttab",)

#: tuple : suffixes that designate inner nested tables in EMu
NESTTAB_INNER_SUFFIXES = ("_nesttab_inner",)

#: str : pattern that matches table suffixes
TAB_PATTERN = "(" + "|".join(TAB_SUFFIXES) + ")$"

#: str : pattern that matches reference suffixes
REF_PATTERN = "(" + "|".join(REF_SUFFIXES) + ")$"

#: str : pattern that matches update modifiers
MOD_PATTERN = r"\(\d*[=\+\-]\)$"


@lru_cache(maxsize=None)
def is_tab(field):
    """Checks if a field name is a table

    Parameters
    ----------
    field : str
        field name

    Returns
    -------
    bool
        True if field name is a table, False if not
    """
    return strip_mod(field).endswith(TAB_SUFFIXES)


@lru_cache(maxsize=None)
def is_nesttab(field):
    """Checks if a field name is a nested table

    Parameters
    ----------
    field : str
        field name

    Returns
    -------
    bool
        True if field name is a nested table, False if not
    """
    return strip_mod(field).endswith(NESTTAB_SUFFIXES)


@lru_cache(maxsize=None)
def is_nesttab_inner(field):
    """Checks if a field name is an inner nested table

    Parameters
    ----------
    field : str
        field name

    Returns
    -------
    bool
        True if field name is an inner nested table, False if not
    """
    return strip_mod(field).endswith(NESTTAB_INNER_SUFFIXES)


@lru_cache(maxsize=None)
def is_ref_tab(field):
    """Checks if a field name is a reference table

    Parameters
    ----------
    field : str
        field name

    Returns
    -------
    bool
        True if field name is a reference table, False if not
    """
    return is_tab(field) and is_ref(field)


@lru_cache(maxsize=None)
def is_ref(field):
    """Checks if a field name is a reference

    Parameters
    ----------
    field : str
        field name

    Returns
    -------
    bool
        True if field name is a reference, False if not
    """
    return strip_mod(field).endswith(REF_SUFFIXES)


@lru_cache(maxsize=None)
def has_mod(field):
    """Checks if a field name ends with an update modifier

    Parameters
    ----------
    field : str
        field name

    Returns
    -------
    bool
        True if field name ends with an update modifier, False if not
    """
    result = bool(re.search(MOD_PATTERN, field))
    if result and not is_tab(field):
        raise ValueError(f"Update modifier found on an atomic field: {field}")
    return result


@lru_cache(maxsize=None)
def strip_tab(field):
    """Strips table suffixes from a field name

    Parameters
    ----------
    field : str
        field name

    Returns
    -------
    str
        field name without a table suffix
    """
    return re.sub(TAB_PATTERN, "", strip_mod(field))


@lru_cache(maxsize=None)
def strip_mod(field):
    """Strips update modifier from a field name

    Parameters
    ----------
    field : str
        field name

    Returns
    -------
    str
        field name without an update modifier
    """
    return field.rsplit("(", 1)[0]


@lru_cache(maxsize=None)
def get_mod(field):
    """Gets the update modifier from a field name

    Parameters
    ----------
    field : str
        field name

    Returns
    -------
    str
        a modifier if found, otherwise an empty string
    """
    if not field.endswith(")"):
        return ""
    mod = "(" + field.rsplit("(", 1)[-1]
    if not re.match(MOD_PATTERN, mod):
        raise ValueError(f"Invalid modifier: {mod}")
    return mod.strip("()")


def flatten(obj, path=None, result=None):
    """Flattens a record to a one-level dict

    Parameters
    ----------
    obj : list
        a list of EMuRecords

    Returns
    -------
    list
        records flattened to one level
    """
    if path is None:
        path = []
        result = {}
    if isinstance(obj, dict):
        for key, val in obj.items():
            path.append(key)
            flatten(val, path, result)
            path.pop()
    elif isinstance(obj, list):
        for i, val in enumerate(obj):
            path.append(f"{i + 1}.{strip_tab(path[-1])}")
            flatten(val, path, result)
            path.pop()
    else:
        result[".".join(path)] = obj
    return result



"""Defines containers to read and write various EMu objects"""

import json
import logging
import os
import pickle
import re
from collections.abc import MutableMapping, MutableSequence
from ctypes import c_uint64
from copy import deepcopy
from functools import lru_cache
from pathlib import Path
from pprint import pformat
from textwrap import wrap
from warnings import warn

from lxml import etree
import yaml

from .io import EMuReader
from .types import EMuDate, EMuFloat, EMuLatitude, EMuLongitude, EMuTime
from .utils import (
    is_ref,
    is_nesttab,
    is_nesttab_inner,
    is_tab,
    get_mod,
    has_mod,
    strip_mod,
    strip_tab,
)


logger = logging.getLogger(__name__)


class EMuConfig(MutableMapping):
    """Reads and writes a configuration file

    Automatically loaded when EMuRecord is first accessed. The current configuration
    can be accessed using the config attribute on each of the EMu classes.

    Parameters
    ----------
    path : str
        path to the config file. If omitted, checks the current and home
        directories for the file.

    Attributes
    ----------
    path : str
        path to the config file
    title : str
        title to write at the top of the config file
    filename : str
        default filename for config file
    classes : list
        list of classes to add the config object to
    """

    def __init__(self, path=None):
        self.path = path
        self.title = "YAML configuration file for python xmu package"
        self.filename = ".xmurc"
        self.classes = [
            EMuSchema,
            EMuReader,
            EMuRecord,
            EMuColumn,
            EMuGrid,
            EMuRow,
        ]
        self._config = None

        # Options as key: (default, comment)
        self._options = {
            "schema_path": (
                "",
                (
                    "Path to a schema.pl file. A JSON copy of the schema will be"
                    " created in the same directory the first time xmu is run."
                ),
            ),
            "groups": (
                {},
                (
                    "Additional groups not defined in the schema. Schema groups"
                    " correspond to grids, so groups that include non-grid fields"
                    " fields (usually tabs, like the Lat/Long tab in Collection"
                    " Events, where the non-grid content changes depending on"
                    " which row is selected) do not include a complete list of"
                    " fields that should be included when updating the group."
                ),
            ),
            "make_visible": (
                [],
                (
                    "Path as 'module.field' to fields missing an ItemName entry in the"
                    " schema that should resolve even if schema.visible_only is True."
                ),
            ),
            "lookup_no_autopopulate": (
                [],
                (
                    "# Path as 'module.field' to fields that should not be populated"
                    " when filling a lookup hierarchy during import."
                ),
            ),
        }

        self.load_rcfile()

        # Set config parameter on all classes
        for cl in self.classes:
            cl.config = self

    def __str__(self):
        return f"{self.__class__.__name__}({pformat(self._config)})"

    def __repr__(self):
        return repr(self._config)

    def __getitem__(self, key):
        return self._config[key]

    def __setitem__(self, key, val):
        self._config[key] = val

    def __delitem__(self, key):
        del self._config[key]

    def __len__(self):
        return len(self._config)

    def __iter__(self):
        return iter(self._config)

    def load_rcfile(self, path=None):
        """Loads a configuration file

        Parameters
        ----------
        path : str
            path to the rcfile. If not given, checks the current then home
            directory for the filename.

        Returns
        -------
        dict
            either a custom configuration loaded from a file or the
            default configuration defined in this function
        """

        if path is None:
            path = self.path

        # Check the current then home directories if path not given
        default_paths = [".", os.path.expanduser("~")]
        paths = default_paths if path is None else [path]

        # Create a default configuration based on _options attribute
        self._config = {k: v[0] for k, v in self._options.items()}

        # Check each location for the rcfile
        for path in paths:
            # Use a default filename if none given
            if os.path.isdir(path):
                path = os.path.join(path, self.filename)

            try:
                with open(path, encoding="utf-8") as f:
                    self.update(yaml.safe_load(f))
                self.path = path
                break
            except FileNotFoundError:
                pass

        return self._config

    def save_rcfile(self, path=None, overwrite=False):
        """Saves a configuration file

        Parameters
        ----------
        path : str
            path for the rcfile. If a directory, adds the filename.
            Defaults to the user's home directory.
        overwrite : bool
            whether to overwrite the file if it exists
        """

        # Default to user home directory
        if path is None:
            path = os.path.expanduser("~")

        # Use a default filename if none given
        if os.path.isdir(path):
            path = os.path.join(path, self.filename)

        # Check if a file already exists at the path
        try:
            with open(path, encoding="utf-8") as f:
                pass
            if overwrite:
                raise FileNotFoundError
            raise IOError(
                f"'{path}' already exists. Use overwrite=True to overwrite it."
            )
        except FileNotFoundError:
            pass

        # Write a commented YAML file. Comments aren't supported by pyyaml
        # and have to be hacked in.
        content = [f"# {self.title}"]
        for line in yaml.dump(self._config, sort_keys=False).splitlines():
            try:
                comment = self._options[line.split(":")[0]][1]
                wrapped = "\n".join([f"# {l}" for l in wrap(comment)])
                content.extend(["", wrapped, line])
            except KeyError:
                # Catches keys that are not top-level options
                content.append(line)

        with open(path, "w", encoding="utf-8") as f:
            f.write("\n".join(content))

    def update(self, obj, path=None):
        """Recusrively updates configuration from dict

        Parameters
        ----------
        obj : mixed
            configuration object. Usually a dict, although the function itself
            may pass a variety of object types.
        path : list
            path to the current item

        Returns
        -------
        None
        """
        if path is None:
            path = []

        if path:
            config = self._config
            for key in path[:-1]:
                config = config.setdefault(key, {})

        if isinstance(obj, dict):
            for key, val in obj.items():
                path.append(key)
                self.update(val, path)
                path.pop()
        elif isinstance(obj, list):
            config[path[-1]] = []
            for i, val in enumerate(obj):
                config[path[-1]].append(type(val)())
                path.append(i)
                self.update(val, path)
                path.pop()
        else:
            if isinstance(obj, str) and obj.startswith("~"):
                obj = os.path.realpath(os.path.expanduser(obj))
            try:
                config[path[-1]] = obj
            except IndexError:
                config.append(obj)


class EMuSchema(dict):
    """Reads and queries an EMu schema file

    Sets the schema class attribute on all classes defined in this module
    when called. If a schema_path is specified in the config file, the
    schema is loaded automatically the first time an EMuRecord is created.

    Parameters
    ----------
    args, kwargs :
        any arguments that can be used to create a dict. If a single string,
        tries to load the dict from the path represented by that string. If
        omitted, will check the config attribute for a schema path.

    Attributes
    ----------
    path : str
        path to the schema file, if used
    visible_only : bool
        whether to resolve fields without an ItemName attribute in the schema
    validate_paths : bool
        whether to validate paths as they are added to an EMuRecord
    """

    #: EMuConfig : module-wide configuration parameters. Set automatically
    #: when an EMuConfig object is created.
    #:
    #: :meta hide-value:
    config = None

    def __init__(self, *args, **kwargs):
        # Load a config file from one of the default locations if empty
        if self.config is None:
            EMuConfig()

        # Disable both checks for the initial read
        self.visible_only = False
        self.validate_paths = False

        if not args or kwargs:
            try:
                args = [self.config["schema_path"]]
            except TypeError:
                pass

        self.path = None
        if len(args) == 1 and isinstance(args[0], (str, Path)):
            self.from_file(args[0])
        elif args or kwargs:
            super().__init__(*args, **kwargs)

        # Enable both checks by default
        self.visible_only = True
        self.validate_paths = True

        # Set schema parameter on all classes
        EMuReader.schema = self
        EMuRecord.schema = self
        EMuColumn.schema = self
        EMuGrid.schema = self
        EMuRow.schema = self

        # Add custom groups from config file. This needs to come after the
        # assignment of the class attributes because _get_field_info() uses
        # one of those to access the schema.
        if self.config is not None:
            for module, groups in self.config["groups"].items():
                for fields in groups.values():
                    self.define_group(module, fields, overwrite=True)

    def __getitem__(self, path):
        path = _split_path(path)
        obj = super().__getitem__(path[0])
        try:
            for key in path[1:]:
                obj = obj[key]
        except KeyError as exc:
            try:
                similar = self._get_similar_keys(path)
            except KeyError:
                raise KeyError(f"Path not found: {path} (failed at {key})") from exc
            else:
                raise KeyError(
                    f"Path not found: {path} (failed at {key}, similar keys include {similar})"
                ) from exc
        return obj

    @property
    def modules(self):
        """Gets the list of modules in the schema"""
        return sorted(self["Schema"].keys())

    def get(self, key, default=None):
        """Overrides the native get method to support paths

        Parameters
        ----------
        key : mixed
            key to retrieve
        default : mixed
           default value to return if key not found

        Returns
        -------
        mixed
            value for the key or default if not found
        """
        try:
            return self[key]
        except KeyError:
            return default

    def from_file(self, path):
        """Loads schema from a file, creating a JSON version if not found

        Parameters
        ----------
        path : str
            path to a schema file
        """
        self.path = path
        path = os.path.splitext(path)[0]
        try:
            self.from_json(f"{path}.json")
        except FileNotFoundError:
            self.from_pl(f"{path}.pl")

            # Map group definitions to fields prior to saving the JSON file
            EMuRecord.schema = self
            for module, data in self["Schema"].items():
                for fields in data.get("groups", {}).values():
                    self.define_group(module, fields)

            self.to_json(f"{path}.json")

    def from_pl(self, path):
        """Loads schema from an EMu schema.pl file

        Parameters
        ----------
        path : str
            path to a schema file
        """
        self.update(self._read_schema_pl(path))

    def from_json(self, path):
        """Loads schema from JSON

        Parameters
        ----------
        path : str
            path to a JSON schema file
        """
        with open(path, encoding="utf-8") as f:
            self.update(json.load(f))

    def to_json(self, path, **kwargs):
        """Saves schema to JSON

        Parameters
        ----------
        path : str
            path to a JSON schema file
        kwargs :
            keyword arguments to pass to json.dump() to control the format
            of the JSON file. Method saves the JSON compactly by default.
        """
        params = {
            "ensure_ascii": False,
            "indent": None,
            "sort_keys": False,
            "separators": (",", ":"),
        }
        params.update(**kwargs)
        with open(path, "w", encoding="utf-8") as f:
            json.dump(self, f, **params)

    def iterfields(self):
        """Iterates over all fields in the schema

        Yields
        ------
        tuple
            (module, field, field info) for each field
        """
        for module in self.modules:
            cols = self[("Schema", module, "columns")]
            for field, info in cols.items():
                yield module, field, info

    def define_group(self, module, fields, overwrite=False):
        """Maps a group definition to each member field

        Groups are read from a schema if possible but can also be defined manually.

        Parameters
        ----------
        module : str
            backend module name
        fields : list
            list of fields in the group
        overwrite : bool
            whether to overwrite an existing group
        """
        for field in fields:
            info = self.get_field_info(module, field)

            # Group definitions in the schema point to the client field that
            # manages a reference (e.g., ModField_tab instead of ModFieldRef_tab).
            # Use the reference table instead.
            try:
                fields[fields.index(field)] = info["RefLink"]
                info = self.get_field_info(module, info["RefLink"])
            except KeyError:
                pass

            if info.get("GroupFields") and not overwrite:
                warn(f"{module}.{field} is assigned to multiple groups")
            else:
                info["GroupFields"] = fields

        # Field definitions modified, so clear the cache
        _get_field_info.cache_clear()

    @staticmethod
    def get_field_info(module, path, visible_only=None):
        """Gets data about the field specified by a path

        Parameters
        ----------
        module : str
            backend module name
        path : str
            path to the field in EMu
        visible_only : bool
            whether to resolve fields that do not appear in the client

        Returns
        -------
        dict
            information about the field (names, data types, etc.)
        """
        return _get_field_info(module, path, visible_only=visible_only)

    def _read_schema_pl(self, path):
        """Reads an EMu schema file

        Parameters
        ----------
        path : str
            path to a schema.pl file

        Returns
        -------
        dict
            EMu schema
        """
        schema = {"Schema": {}}
        dct = schema["Schema"]
        keypath = ["Schema"]
        items = []
        with open(path, encoding="utf-8") as f:
            for line in f:
                line = line.strip().strip(",")

                # Open new child dictionary when line ends with an arrow
                if line.endswith("=>"):
                    key = line.rsplit("=>")[0].strip().strip("'")
                    keypath.append(key)
                    dct[key] = {}
                    dct = dct[key]

                # Add new key-val when arrow occurs mid-line
                elif "=>" in line:
                    key, val = [s.strip().strip("'") for s in line.split("=>")]
                    dct[key] = self._parse_value(val)

                # Open a new list
                elif line == "[":
                    items.append([])

                # Set list as the value for the current key
                elif line == "]" and len(items) == 1:
                    last = keypath.pop(-1)
                    dct = schema
                    for key in keypath:
                        dct = dct[key]
                    dct[last] = items.pop(-1)

                # Append nested list to previous item in the lists container
                elif line == "]":
                    items[-2].append(items.pop(-1))

                # Append value to the current list
                elif items:
                    vals = self._parse_value(line)
                    if line.startswith("[") or isinstance(vals, (str, int)):
                        items[-1].append(vals)
                    else:
                        items[-1].extend(vals)

                # Go up one level in the dictionary
                elif line == "}":
                    keypath.pop(-1)
                    dct = schema
                    for key in keypath:
                        dct = dct[key]

        return schema

    def _get_similar_keys(self, path):
        """Finds fields similar to the one at the end of the given path"""
        last = path[-1][:4]
        obj = super().__getitem__(path[0])
        for key in path[1:-1]:
            obj = obj[key]
        return [k for k in obj if k.startswith(last)]

    @staticmethod
    def _parse_value(val):
        """Parses strings, lists, and integers from perl file"""
        vals = val.strip("[]").split(",")
        try:
            vals = [int(s.strip()) for s in vals]
        except ValueError:
            vals = [s.strip("'").strip() for s in vals]
        return vals if "," in val else vals[0]


class EMuColumn(list):
    """Reads and writes data in a table field

    Parameters
    ----------
    vals : iterable
        values for the column
    module : str
        backend name of an EMu module
    field : str
        name of an EMu field
    dict_class : EMuRecord
        class to use for dicts
    list_class : EMuColumn
        class to use for lists

    Attributes
    ----------
    module : str
        backend name of an EMu module
    field : str
        name of an EMu field
    dict_class : EMuRecord
        class to use for dicts
    list_class : EMuColumn
        class to use for lists
    """

    #: EMuConfig : module-wide configuration parameters. Set automatically
    #: when an EMuConfig object is created.
    #:
    #: :meta hide-value:
    config = None

    #: EMuSchema : info about a specific EMu configuration. Set automatically
    #: when an EMuSchema object is created.
    #:
    #: :meta hide-value:
    schema = None

    def __init__(
        self, vals=None, module=None, field=None, dict_class=None, list_class=None
    ):
        self.module = module
        self.field = field
        self.dict_class = dict_class if dict_class else DEFAULT_RECORD
        self.list_class = list_class if list_class else DEFAULT_COLUMN
        if self.schema and not self.module:
            raise ValueError(
                f"Must provide module when schema is used (one of {self.schema.modules})"
            )
        super().__init__()
        if vals:
            self.extend(vals)

    def __str__(self):
        return f"EMuColumn({super().__str__()})"

    def __setitem__(self, i, val):
        # Catch non-integer indices to avoid problems with coercing values
        if not isinstance(i, int):
            raise TypeError("list indices must be integers or slices, not str")
        super().__setitem__(i, _coerce_values(self, val))

    def __add__(self, obj):
        return self.__class__(
            super().__add__(obj), module=self.module, field=self.field
        )

    def __iadd__(self, obj):
        self.extend(obj)
        return self

    def insert(self, i, val):
        super().insert(i, _coerce_values(self, val))

    def append(self, val):
        super().append(_coerce_values(self, val))

    def extend(self, vals):
        super().extend([_coerce_values(self, v) for v in vals])

    def copy(self):
        """Overrides the native list.copy method to return an object of this class"""
        return pickle.loads(pickle.dumps(self))

    def to_xml(self, root=None, kind=None, row_ids=None):
        """Converts column to XML formatted for EMu

        Normally called without specifying arguments.

        Parameters
        ----------
        root : lxml.etree.Element or SubElement
            parent element in the XML tree
        kind : str
           kind of XML file. One of "import", "update", or "emu".
        row_ids : tuple
            list of values for the tuple group attribute

        Returns
        -------
        lxml.etree.Element or SubElement
            table as XML
        """
        name = strip_mod(self.field)
        mod = get_mod(self.field)

        if root is None:
            root = etree.Element("table")
            root.set("name", name)
        elif root.tag != "table":
            root = etree.SubElement(root, "table")
            root.set("name", name)

        for (
            i,
            child,
        ) in enumerate(self):
            tup = etree.SubElement(root, "tuple")

            # Add group and row indicators for updates
            if mod:
                tup.set("row", mod)
                if row_ids and mod == "+":
                    tup.set("group", row_ids[i])
            # Otherwise explicitly number the table rows. This prevents EMu from
            # skipping empty nested table cells when reading an import file.
            else:
                tup.set("row", str(i + 1))

            try:
                child.to_xml(tup, kind=kind)
            except AttributeError:
                # In an EMu export, empty rows in an outer nested table appear
                # as an empty tuple. Otherwise tuples always contain one or more
                # atomic fields, including an empty irn field for references.
                if _is_not_blank(child) or not is_nesttab(self.field):
                    # Interpret an atomic value inside a reference table as an irn
                    name = "irn" if is_ref(self.field) else strip_tab(self.field)
                    atom = etree.SubElement(tup, "atom")
                    atom.set("name", name)

                    # Set text, deferring to the emu_str method if it exists
                    atom.text = ""
                    if _is_not_blank(child):
                        try:
                            atom.text = child.emu_str()
                        except AttributeError:
                            atom.text = str(child)

        return root


class EMuRow(MutableMapping):
    """Reads and writes data in a grid row

    Changes to the row are reflected in the original EMuRecord

    Parameters
    ----------
    rec : XMuRecord
        the EMu record the grid is from
    path : str
        path to a field that is part of the grid
    index : int
        the index of the row
    fill_value : mixed
        value used when deleting an item from the row

    Attributes
    ----------
    group : tuple
        names for all columns that are part of the parent grid, whether they appear
        in the current record or not
    fill_value : mixed
        value used when deleting an item from the row
    """

    #: EMuConfig : module-wide configuration parameters. Set automatically
    #: when an EMuConfig object is created.
    #:
    #: :meta hide-value:
    config = None

    #: EMuSchema : info about a specific EMu configuration. Set automatically
    #: when an EMuSchema object is created.
    #:
    #: :meta hide-value:
    schema = None

    def __init__(self, rec, path, index, fill_value=None):
        module = _get_module(rec)
        self.group = tuple(
            self.schema.get_field_info(module, path).get("GroupFields", [])
        )
        if not self.group:
            raise KeyError(f"{module}.{path} is not part of a group")
        self.fill_value = fill_value
        self.index = index

        # Use path to drill down to the correct parent record
        path = _split_path(path)[:-1]
        if path:
            rec = rec[path]
        self._rec = rec

    def __str__(self):
        try:
            row = {c: self._rec[c][self.index] for c in self.columns}
        except IndexError:
            raise IndexError(
                "One or more columns has no data for this row. Use pad() on the parent grid to prevent this error."
            )
        return f"{self.__class__.__name__}({row})"

    def __repr__(self):
        return str(self)

    def __iter__(self):
        return iter(self.columns)

    def __len__(self):
        return len(self.columns)

    def __setitem__(self, key, val):
        self._rec[key][self.index] = val

    def __getitem__(self, key):
        return self._rec[key][self.index]

    def __delitem__(self, key):
        self[key] = self.fill_value

    @property
    def columns(self):
        """Lists columns in the row that exist in the record"""
        cols = [c for c in self._rec if strip_mod(c) in set(self.group)]
        if (
            any((c.endswith(")") for c in cols))
            and len({c.rsplit("(", 1)[-1] for c in cols}) > 1
        ):
            raise ValueError(f"Inconsistent modifier within grid: {cols}")
        return cols

    @property
    def replace_mod(self):
        """Returns the modifier needed to replace a cell in this row in an import"""
        return f"{self.index + 1}="

    def row_id(self):
        """Calculates an identifier based on the index and content of a row"""
        val = str(self.index) + str(self)
        return c_uint64(hash(val)).value.to_bytes(8, "big").hex()


class EMuGrid(MutableSequence):
    """Reads and writes data in a grid

    Changes to the grid are reflected in the original EMuRecord

    Parameters
    ----------
    rec : XMuRecord
        the EMu record the grid is from
    path : str
        path to a field that is part of the grid
    fill_value : mixed
        value used when padding columns
    pad : bool
        whether to pad the columns to the same length

    Attributes
    ----------
    group : tuple
        names for all columns that are part of this grid, whether they appear
        in the current record or not
    fill_value : mixed
        value to use when padding the grid or deleting an item from an EMuRow
        object created from this grid
    """

    #: EMuConfig : module-wide configuration parameters. Set automatically
    #: when an EMuConfig object is created.
    #:
    #: :meta hide-value:
    config = None

    #: EMuSchema : info about a specific EMu configuration. Set automatically
    #: when an EMuSchema object is created.
    #:
    #: :meta hide-value:
    schema = None

    def __init__(self, rec, path, fill_value=None):
        module = _get_module(rec)
        self.group = tuple(
            self.schema.get_field_info(module, path).get("GroupFields", [])
        )
        if not self.group:
            raise KeyError(f"{module}.{path} is not part of a group")
        self.fill_value = fill_value

        # Use path to drill down to the correct parent record
        path = _split_path(path)[:-1]
        if path:
            rec = rec[path]
        self._rec = rec

    def __str__(self):
        return f"{self.__class__.__name__}({list(self)})"

    def __repr__(self):
        return str(self)

    def __iter__(self):
        for i in range(len(self)):
            yield EMuRow(self._rec, self.columns[0], i, fill_value=self.fill_value)

    def __len__(self):
        try:
            return max([len(self._rec[c]) for c in self.columns])
        except ValueError:
            return 0

    def __getitem__(self, key):
        if isinstance(key, int):
            return list(self)[key]

        if isinstance(key, dict):
            matches = []
            for row in self:
                for key_, val in key.items():
                    try:
                        row_val = self._transform(row.get(key_))
                    except IndexError:
                        raise IndexError(
                            f"Key in query does not appear in all rows: {repr(key_)}"
                        )
                    match_val = self._transform(val)
                    # Match if values are equal or if row_val matches one of
                    # multiple options given for match_val
                    if row_val == match_val or (
                        row_val and match_val and row_val in match_val.split("|")
                    ):
                        continue
                    break
                else:
                    matches.append(row)
            return matches

        return self._rec[key]

    def __setitem__(self, i, vals):
        # Required by MutableSequence but does not make sense to implement
        raise NotImplementedError(
            "Cannot set items on an EMuGrid. Use the main EMuRecord object or"
            " an individual EMuRow instead."
        )

    def __delitem__(self, i):
        for col in self.columns:
            del self._rec[col][i]

    def __eq__(self, other):
        return list(self) == list(other)

    @property
    def columns(self):
        """Lists columns in the grid that exist in the record"""
        cols = [c for c in self._rec if strip_mod(c) in set(self.group)]
        if (
            any((c.endswith(")") for c in cols))
            and len({c.rsplit("(", 1)[-1] for c in cols}) > 1
        ):
            raise ValueError(f"Inconsistent modifier within grid: {cols}")
        return cols

    def verify(self):
        """Checks if any fields in the grid are missing from the record"""
        missing = set(self.group) - set(self.columns)
        if missing:
            raise ValueError(
                f"Grid including '{self.columns[0]}' is missing fields: {missing}."
                f" Verify that all grid fields are present in import, then run"
                f" grid.add_columns().pad() to complete the grid."
            )

    def insert(self, index, value):
        # Required by MutableSequence but does not make sense to implement
        raise NotImplementedError(
            "Cannot insert into an EMuGrid. Use the main EMuRecord object instead."
        )

    def items(self):
        for col in self.columns:
            yield col, self._rec[col]

    def add_columns(self, cols=None, fill_value=None):
        """Adds missing columns to the grid

        Parameters
        ----------
        cols : list-like
            columns to add. If not given, adds all columns in the group attribute
            that do not already appear in the record.
        fill_value : mixed
            the value used to pad a column. Defaults to fill_value attribute
            if not given.

        Returns
        -------
        self
        """
        if fill_value is None:
            fill_value = self.fill_value
        mod = get_mod(self.columns[0]) if self.columns else None
        if cols is None:
            cols = self.group
        if mod:
            cols = [f"{c}({mod})" if not has_mod(c) else c for c in cols]
        for col in set(cols) - set(self.columns):
            self._rec.setdefault(col, [fill_value for _ in range(len(self))])
        return self

    def pad(self, fill_value=None):
        """Pads all columns in the table to the same length

        Parameters
        ----------
        fill_value : mixed
            the value used to pad a column. Defaults to fill_value attribute
            if not given.

        Returns
        -------
        self
        """
        if fill_value is None:
            fill_value = self.fill_value
        for col in self.columns:
            diff = len(self) - len(self._rec[col])
            self._rec[col].extend([fill_value for _ in range(diff)])
        return self

    def query(self, field=None, where=None):
        """Queries the grid

        Parameters
        ----------
        field : str
            a specific column to return. If empty, the whole row is returned.
        where : dict
            the query as a dict. A row must match all criteria to be returned.

        Returns
        -------
        list
            list of matching rows or values
        """
        results = []
        for row in self[where]:
            results.append(row[field] if field else row)
        return results

    @staticmethod
    def _transform(val):
        if not isinstance(val, (list, tuple)):
            val = [val]
        return "|".join([str(s) if s is not None else "" for s in val]).lower()


class EMuRecord(dict):
    """Reads and writes data in an EMu record

    Parameters
    ----------
    rec : mapping or iterable
        record as a mapping or iterable
    module : str
        backend name of an EMu module
    field : str
        name of an EMu field
    dict_class : EMuRecord
        class to use for dicts
    list_class : EMuColumn
        class to use for lists

    Attributes
    ----------
    module : str
        backend name of an EMu module
    field : str
        name of an EMu field
    dict_class : EMuRecord
        class to use for dicts
    list_class : EMuColumn
        class to use for lists
    """

    #: EMuConfig : module-wide configuration parameters. Set automatically
    #: when an EMuConfig object is created.
    #:
    #: :meta hide-value:
    config = None

    #: EMuSchema : info about a specific EMu configuration. Set automatically
    #: when an EMuSchema object is created.
    #:
    #: :meta hide-value:
    schema = None

    def __init__(
        self, rec=None, module=None, field=None, dict_class=None, list_class=None
    ):
        self.module = module
        self.field = field
        self.dict_class = dict_class if dict_class else DEFAULT_RECORD
        self.list_class = list_class if list_class else DEFAULT_COLUMN

        if self.schema and not self.module:
            raise ValueError(
                f"Must provide module when schema is used (one of {self.schema.modules})"
            )

        super().__init__()
        if rec:
            if isinstance(rec, str):
                rec = json.loads(rec)
            elif isinstance(rec, self.__class__):
                rec = rec.copy()
            self.update(rec)

    def __str__(self):
        return f"{self.__class__.__name__}({pformat(self)})"

    def __getitem__(self, path):
        path = _split_path(path)
        try:
            if len(path) > 1:
                obj = self
                for key in path:
                    try:
                        obj = obj[key]
                    except TypeError:
                        if obj is None:
                            raise KeyError
                        raise
                return obj
            key = path[0]
            return super().__getitem__(key)
        except KeyError as exc:
            # Check path against schema if key not found
            module = _get_module(self)
            dotpath = ".".join(path)
            if module and self.schema is not None and self.schema.validate_paths:
                try:
                    self.schema.get_field_info(module, path)
                except KeyError:
                    warn(f"Invalid path: {dotpath} (module={module})")
                    raise KeyError(
                        f"Invalid path: {dotpath} (module={module})"
                    ) from exc
                else:
                    raise KeyError(
                        f"Path not found but valid: {dotpath} (module={module})"
                    ) from exc
            raise KeyError(
                f"Path not found: {dotpath} (module={module}) (failed at {key})"
            ) from exc

    def __setitem__(self, key, val):
        # Catch a key containing illegal characters
        if not re.match(r"\w+$", strip_mod(key)):
            raise ValueError(f"Invalid key: {key} (module={_get_module(self)})")
        super().__setitem__(key, _coerce_values(self, val, key))

    def get(self, key, default=None):
        """Overrides the native dict.get method to map unrecognized terms"""
        try:
            return self[key]
        except KeyError:
            return default

    def setdefault(self, key, val):
        """Overrides the native dict.setdefault method to use the subclass setter"""
        try:
            return self[key]
        except KeyError:
            self[key] = val
            return self[key]

    def update(self, *args, **kwargs):
        """Overrides the native dict.update method to use the subclass setter"""
        for key, val in dict(*args, **kwargs).items():
            self[key] = val

    def copy(self):
        """Overrides the native dict.copy method to return an object of this class"""
        return pickle.loads(pickle.dumps(self))

    def json(self, **kwargs):
        """Converts record to JSON

        Parameters
        ----------
        kwargs :
            any kwarg accepted by json.dumps

        Returns
        -------
        str
            record as JSON string
        """
        kwargs.setdefault("cls", EMuEncoder)
        kwargs.setdefault("ensure_ascii", False)
        return json.dumps(dict(self), **kwargs)

    def grid(self, field, **kwargs):
        """Returns the EMuGrid object containing the given field

        Parameters
        ----------
        field : str
            any field name that appears in a grid
        kwargs :
            keyword agruments for EMuGrid

        Returns
        -------
        EMuGrid
            the grid from the current record
        """
        return EMuGrid(self, field, **kwargs)

    def to_xml(self, root=None, kind=None):
        """Converts record to XML formatted for EMu

        Normally called without specifying arguments.

        Parameters
        ----------
        root : lxml.etree.Element or SubElement
            parent element in the XML tree
        kind : str
           kind of XML file. One of "import", "update", or "emu". If not
           given, assigns "update" if top-level records have irns and "import"
           if not.

        Returns
        -------
        lxml.etree.Element or SubElement
            record as XML
        """
        kinds = {None, "emu", "import", "update"}
        if kind not in {None, "emu", "import", "update"}:
            raise ValueError(f"kind must be one of {kinds}")

        if root is None:
            root = etree.Element("tuple")
        elif root.get("name") == self.module:
            root = etree.SubElement(root, "tuple")

        # Records containing irns in the top level of the dict are updates
        if kind is None:
            kind = "update" if "irn" in self else "import"

        # Fill in grids and cache row IDs so grids are only checked once
        grids = {}
        for key in list(self):
            if kind == "update":
                try:
                    grids[key]
                except KeyError:
                    try:
                        grid = self.grid(key)
                    except KeyError:
                        pass
                    else:
                        # Include all columns when appending or prepending
                        if key.endswith(("(+)", "(-)")):
                            grid.add_columns()
                        grid.pad()
                        row_ids = [r.row_id() for r in grid]
                        for col in grid.columns:
                            grids[col] = row_ids

            # Populate unfilled lookup list parent fields
            module = _get_module(self)
            if self.schema is not None and self.schema.validate_paths:
                while True:
                    field_info = self.schema.get_field_info(module, key)
                    try:
                        lookup_parent = field_info["LookupParent"]
                    except KeyError:
                        break

                    # Break on SecLookupRoot and any other fields specified in config
                    if (
                        lookup_parent == "SecLookupRoot"
                        or f"{self.module}.{lookup_parent}"
                        not in self.config["lookup_no_autopopulate"]
                    ):
                        break

                    try:
                        self[lookup_parent]
                    except KeyError:
                        logger.debug(f"Filled parent in lookup: {lookup_parent}")
                        self[lookup_parent] = None

                    key = lookup_parent

        for key, val in self.items():
            if is_tab(key):
                # If field is part of a grid, pass row identifiers to the
                # EMuColumn to_xml() method. These will be used to populate the
                # group attribute in each tuple tag for appends and prepends.
                val.to_xml(root, kind=kind, row_ids=grids.get(key, None))
            elif is_ref(key):
                if isinstance(val, int):
                    # The module does not matter here, so just use the parent's
                    val = self.__class__({"irn": val}, module=self.module)
                ref_tup = etree.SubElement(root, "tuple")
                ref_tup.set("name", key)
                if val is not None:
                    val.to_xml(ref_tup, kind=kind)
            else:
                atom = etree.SubElement(root, "atom")
                atom.set("name", key)

                # Set text, deferring to the emu_str method if it exists
                atom.text = ""
                if _is_not_blank(val):
                    try:
                        atom.text = val.emu_str()
                    except AttributeError:
                        atom.text = str(val)
        return root


class EMuEncoder(json.JSONEncoder):
    """Encodes objects using EMuRecord and EMuColumn"""

    def default(self, obj):
        if isinstance(obj, dict):
            return dict(obj)
        elif isinstance(obj, list):
            return list(obj)
        return str(obj)


def _coerce_values(parent, child, key=None):
    """Coerces child containers and values to specific classes"""

    # Pickled objects are missing the instance attributes required to
    # use this function, but since these objects have already been
    # coerced, they can be returned as is.
    try:
        parent.module
    except AttributeError:
        return child

    if isinstance(parent, dict):
        dict_class = parent.dict_class
        list_class = parent.list_class
        field = key
        module = _get_module(parent)

    # List items inherit the field attribute from their parent
    elif isinstance(parent, (list, tuple)):
        dict_class = parent.dict_class
        list_class = parent.list_class
        field = parent.field
        module = parent.module

    # Validate field if schema has been loaded
    field_info = None
    if parent.schema and parent.schema.validate_paths:
        field_info = parent.schema.get_field_info(module, key if key else field)

    # Label inner nested tables
    if is_nesttab(field) and not isinstance(parent, dict_class):
        field = f"{strip_mod(field)}_inner"

    # Tables must be list-like
    if field != parent.field and is_tab(field) and not isinstance(child, (list, tuple)):
        if child is None:
            child = []
        else:
            raise TypeError(
                f"Columns must be lists ({repr(child)} was assigned to {field})"
            )

    # Simplify IRN-only references
    if is_ref(field):
        # Simplify IRN-only references to integers
        if isinstance(child, dict) and list(child) == ["irn"]:
            child = child["irn"]

        # Interpret integers in reference fields as IRNs
        try:
            child_as_int = int(child)
            if isinstance(child_as_int, int):
                return child_as_int
        except (TypeError, ValueError):
            pass

    # References must be dicts or ints (which are interpreted as IRNs)
    if (
        is_ref(field)
        and not is_tab(field)
        and not isinstance(child, dict)
        and child is not None
    ):
        raise TypeError(
            f"References must be dicts ({repr(child)} was assigned to {field})"
        )

    # Sequences must only be used in tables
    if isinstance(child, (list, tuple)) and not is_tab(field):
        raise TypeError(
            f"Sequence assigned to atomic field ({repr(child)} was assigned to {field})"
        )

    # References should use the target module and drop the field
    if isinstance(child, dict) and not isinstance(child, dict_class):
        child = dict_class(child, module=_get_module(parent, field), field=None)

    # Coerce columns to given list class
    elif isinstance(child, (list, tuple)) and not isinstance(child, list_class):
        child = list_class(child, module=module, field=field)

    # Coerce non-list, non-dict data to an appropriate type if a schema is defined
    elif field_info and not isinstance(child, (dict, list)):
        # Coerce common NAs to None
        if (
            isinstance(child, float)
            and str(child) == "nan"
            or str(child) in ("<NA>", "NaT")
        ):
            child = None

        # Coerce empty values to empty strings in Text fields. Exclude
        # inner nested tables so that empty rows can be signified by None.
        dtype = field_info["DataType"]
        if (
            dtype in ("Text", "String")
            and child is None
            and not is_nesttab_inner(field)
        ):
            child = ""

        elif child is not None:
            try:
                child = {
                    "Currency": str,
                    "Date": EMuDate,
                    "Float": EMuFloat,
                    "Integer": int,
                    "Latitude": EMuLatitude,
                    "Longitude": EMuLongitude,
                    "String": str,
                    "Text": str,
                    "Time": EMuTime,
                    "UserName": str,
                    "UserId": str,
                }[dtype](child)
            except (TypeError, ValueError) as exc:
                # Handle integers with decimals or commas
                if (
                    dtype == "Integer"
                    and isinstance(child, str)
                    and (
                        re.search(r"\.0+$", child)
                        or re.match(r"-?\d{1,3}(,\d{3})+", child)
                    )
                ):
                    child = int(EMuFloat(child))
                else:
                    raise TypeError(
                        f"Could not coerce to {dtype} ({field}={repr(child)})"
                    ) from exc

    # Evaluate nesting within tables
    if (
        isinstance(parent, dict)
        and is_tab(field)
        and not is_nesttab_inner(field)
        and child
    ):
        if is_nesttab(field):
            if any((not isinstance(c, list) for c in child if c is not None)):
                raise TypeError(f"Too few levels in a nested table ({field})")
            elif any(
                (
                    any((isinstance(c, list) for c in c if c is not None))
                    for c in child
                    if c is not None
                )
            ):
                raise TypeError(f"Too many levels in a nested table ({field})")
        elif any((isinstance(c, list) for c in child)):
            raise TypeError(f"Too many levels in a table ({field})")

    return child


@lru_cache(maxsize=None)
def _get_field_info(module, path, visible_only=None):
    """Gets field info from a schema for a given module and path

    Moved outside of EMuSchema to allow use of lru_cache.
    """
    schema = EMuRecord.schema

    if visible_only is None:
        visible_only = schema.visible_only

    segments = _split_path(path)
    modules = [module]
    for seg in segments:
        obj = schema[
            ("Schema", modules[-1], "columns", strip_mod(seg).replace("_inner", ""))
        ]
        modules.append(obj.get("RefTable", modules[-1]))

    # ItemName *appears* to be populated only for fields that appear in the client
    module = modules[-2]
    if (
        visible_only
        and not obj.get("ItemName")
        and not ".".join([module] + list(segments)) in EMuRecord.config["make_visible"]
    ):
        raise KeyError(f"{module}.{seg} is valid but not not visible")

    return obj


def _get_module(obj, field=None):
    """Gets module name"""
    if field is None:
        field = obj.field
    if obj.schema is not None and field is not None and is_ref(field):
        return obj.schema.get_field_info(obj.module, field)["RefTable"]
    return obj.module


def _is_not_blank(val):
    """Tests if value is not blank"""
    return val or val == 0


def _split_path(path):
    """Splits path into segments"""
    if isinstance(path, tuple):
        return path
    if isinstance(path, str):
        return tuple(re.split(r"[./]", path))
    if isinstance(path, list):
        return tuple(path)
    raise ValueError(f"Invalid path format: {path}")


DEFAULT_RECORD = EMuRecord
DEFAULT_COLUMN = EMuColumn

# Load config and schema, if provided
try:
    EMuSchema()
except (FileNotFoundError, TypeError):
    pass



import os
from setuptools import setup, find_packages


long_description = (
    "xmu is a Python utility used to read and write XML for Axiell EMu,"
    " a collections management system used in museums, galleries, and"
    " similar institutions."
    "\n\n"
    " Install with:"
    "\n\n"
    "```\n"
    "pip install xmu\n"
    "```"
    "\n\n"
    "Learn more:\n\n"
    "+ [GitHub repsository](https://github.com/NMNH-IDSC/xmu)\n"
    "+ [Documentation](https://xmu.readthedocs.io/en/latest/)"
)


setup(
    name="xmu",
    maintainer="Adam Mansur",
    maintainer_email="mansura@si.edu",
    description="Reads and writes XML for Axiell EMu",
    long_description=long_description,
    long_description_content_type="text/markdown",
    version="0.1b7",
    classifiers=[
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3 :: Only",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3.12",
        "Topic :: Text Processing :: Markup :: XML",
    ],
    url="https://github.com/NMNH-IDSC/xmu.git",
    license="MIT",
    packages=find_packages(),
    install_requires=["joblib", "lxml", "pyyaml"],
    include_package_data=True,
    zip_safe=False,
)



"""Defines objects used to read and write XML for Axiell EMu"""

import csv
import datetime as dt
import glob
import json
import logging
import mmap
import os
import re
import shutil
import tempfile
import time
import zipfile

from joblib import Parallel, delayed
from lxml import etree

from .utils import flatten, is_nesttab, is_ref, is_ref_tab, is_tab, strip_tab


logger = logging.getLogger(__name__)


class EMuReader:
    """Read records from an EMu XML file into dicts

    Parameters
    ----------
    path : str or Path
        path to a file or directory
    json_path : str or Path
        path to a JSON file used to cache records for faster reading

    Attributes
    ----------
    path : str or Path
        path to a file or directory
    json_path : str or Path
        path to a JSON file used to cache records for faster reading
    files : list
        list of file-like objects, each of which is an EMu XML file
    module : str
        the name of an EMu module
    """

    #: EMuConfig : module-wide configuration parameters. Set automatically
    #: when an EMuConfig object is created.
    #:
    #: :meta hide-value:
    config = None

    #: EMuSchema : info about a specific EMu configuration. Set automatically
    #: when an EMuSchema object is created.
    #:
    #: :meta hide-value:
    schema = None

    def __init__(self, path, json_path=None):
        self.path = str(path)
        self._rec_class = dict
        self.json_path = json_path
        self.files = []
        self.module = None
        self._fields = None
        self._get_files()
        self._load_schema()

        # Private attributes used to display notifications
        self._job_start = None
        self._job_done = False
        self._notify_start = None
        self._notify_count = 0

    def __iter__(self):
        for rec in self.from_file():
            yield rec

    def __len__(self):
        counts = self.counts()
        if isinstance(counts, int):
            return counts
        raise NotImplementedError("Not implemented when multiple source files provided")

    @property
    def fields(self):
        if self._fields is None:
            self._fields = self._parse_file_schema()
        return self._fields

    def from_file(self):
        """Reads data from file, using JSON if possible

        Yields
        ------
        dict
            EMu record
        """
        if not self.json_path:
            return self.from_xml()

        # If JSON is older than the newest XML file, regenerate it
        try:
            if os.path.getmtime(self.json_path) < self.files[-1].getmtime():
                logger.info("Regenerating JSON (XML is newer)")
                self.to_json()
        except FileNotFoundError:
            logger.info("Generating JSON (JSON not found)")
            self.to_json()

        return self.from_json()

    def from_xml(self, start=0, limit=None):
        """Reads data from XML

        Parameters
        ----------
        start : int
            index of record to start processing
        limit : int
            number of records to process from start. If omitted, all records
            are processed.

        Yields
        ------
        dict
            EMu record
        """
        try:
            for filelike in self.files:
                logger.info("Reading records from %s", filelike)
                self._job_start = None
                self._job_done = False
                self._notify_start = None
                self._notify_count = 0
                with filelike.open("rb") as source:
                    try:
                        context = etree.iterparse(source, events=["end"], tag="tuple")
                        for _, element in context:
                            # Process children of module table only
                            parent = element.getparent().get("name")
                            if parent is not None and parent.startswith("e"):
                                try:
                                    if self._notify_count >= start:
                                        yield self._parse(element)
                                finally:
                                    element.clear()
                                    # while element.getprevious() is not None:
                                    #    del element.getparent()[0]
                                    self._notify_count += 1
                                    if not self._notify_count % 5000:
                                        logger.info(
                                            "Read %s records from %s",
                                            self._notify_count,
                                            filelike,
                                        )
                                    if (
                                        limit is not None
                                        and self._notify_count >= start + limit
                                    ):
                                        break
                    finally:
                        del context
                logger.info("Read %s records total", self._notify_count)
                self._job_done = True
        finally:
            if self._job_start:
                self.report_progress()

    def from_xml_parallel(
        self, callback, num_parts=64, handle_repeated_keys="overwrite"
    ):
        """Reads data from XML in parallel

        Experimental. Works by creating temporary copies of the XML file, then reading from
        those files in parallel. Seems to work best with a small number of copies.

        Parameters
        ----------
        callback : function
            function to run on the import file
        num_parts : int
            number of parts to split the file into
        handle_repeated_keys : str
            defines how to handle keys that repeat across dicts returned by different jobs.
            Must be one of 'combine' (which combines entires in a list), 'keep' (which keeps
            the first key found), 'overwrite' (which overwrites the existing key), or 'raise'
            (which raises a KeyError). Ignored if callback does not return a dict.

        Yields
        ------
        mixed
            result of callback function combined across jobs. If dict, results are combined
            into a single dict. If list, results are combined into a single list. If another
            type, returns a list of results returned by the callback.
        """

        if len(self.files) > 1 or not self.files[0].path.lower().endswith(".xml"):
            raise NotImplementedError(
                "Not implemented when multiple source files provided"
            )

        allowed = ("combine", "keep", "overwrite", "raise")
        if handle_repeated_keys not in allowed:
            raise ValueError(f"dict_behvaior must be one of the following: {allowed}")

        # Create temporary directory
        tmpdir = tempfile.mkdtemp(prefix="xmu-")

        try:

            files = []
            with open(self.files[0].path, "rb") as f:
                with mmap.mmap(
                    f.fileno(), length=0, access=mmap.ACCESS_READ, offset=0
                ) as m:
                    content = m.read()
                    sep = b"<!-- Row"
                    records = content.split(sep)
                    header = records.pop(0)
                    step = int(len(records) / num_parts) + 1
                    for i in range(0, len(records), step):
                        group = records[i : i + step]
                        tmp = tempfile.NamedTemporaryFile(
                            "wb", prefix="xmu-", suffix=".xml", dir=tmpdir, delete=False
                        )
                        with open(tmp.name, "wb") as f:
                            f.write(header)
                            f.write(b"".join((sep + r for r in group)))
                            if not group[-1].rstrip().endswith(b"</table>"):
                                f.write(b"\n</table>")
                        files.append(tmp)

            results = Parallel(n_jobs=-1)(delayed(callback)(tmp.name) for tmp in files)

            if isinstance(results[0], dict):
                result = {}
                for result_ in results:
                    if handle_repeated_keys == "combine":
                        for key, val in result_.items():
                            if not isinstance(val, list):
                                val = [val]
                            result.setdefault(key, []).extend(val)
                    elif handle_repeated_keys == "keep":
                        for key, val in result_.items():
                            result.setdefault(key, val)
                    elif handle_repeated_keys == "overwrite":
                        result.update(result_)
                    else:
                        if set(result_) & set(result):
                            raise KeyError("Duplicate keys returned")
                        result.update(result_)
                return result
            elif isinstance(results[0], list):
                result = []
                for result_ in results:
                    result.extend(result_)
                return result

            return results
        finally:
            for tmp in files:
                tmp.close()
                os.remove(tmp.name)
            shutil.rmtree(tmpdir)

    def from_json(self, chunk_size=2097152):
        """Reads data from JSON

        Parameters
        ----------
        chunk_size : int
            size of chunk to use when reading the file

        Yields
        ------
        dict
            EMu record
        """
        logger.info("Reading records from %s", self.json_path)
        self._job_start = None
        self._job_done = False
        self._notify_start = None
        self._notify_count = 0
        with open(self.json_path, encoding="utf-8") as f:
            f.read(1)
            add_to_next_chunk = []
            while True:
                chunk = f.read(chunk_size)
                if add_to_next_chunk:
                    chunk = "".join(add_to_next_chunk[::-1]).lstrip(",") + chunk
                    add_to_next_chunk = []

                if len(chunk) <= 1:
                    break

                while True:
                    try:
                        for rec in json.loads(f"[{chunk.lstrip(',')[:-1]}]"):
                            try:
                                yield rec
                            finally:
                                self._notify_count += 1
                                if not self._notify_count % 5000:
                                    logger.info(
                                        "Read %s records from %s",
                                        self._notify_count,
                                        self.json_path,
                                    )
                        break
                    except json.JSONDecodeError:
                        chunk, trailer = chunk.rsplit("{", 1)
                        add_to_next_chunk.append(f"{{{trailer}")
        logger.info("Read %s records total", self._notify_count)
        self._job_done = True
        if self._job_start:
            self.report_progress()

    def to_csv(self, path, **kwargs):
        """Writes records in reader object to CSV

        Parameters
        ----------
        path : str
            path to write the CSV file
        kwargs :
            any keyword argument accepted by open()
        """
        return write_csv(self, path, **kwargs)

    def to_json(self, path=None, **kwargs):
        """Writes JSON version of XML to file

        Parameters
        ----------
        path : str
            path to write JSON
        kwargs :
            keyword arguments for json.dump()
        """
        if path is None:
            path = self.json_path

        logger.info("Writing records from %s to JSON", self.path)

        params = {
            "ensure_ascii": False,
            "indent": None,
            "sort_keys": False,
            "separators": (",", ":"),
        }
        params.update(**kwargs)

        sep = params["separators"][0]

        with open(path, "w", encoding="utf-8"):
            pass

        try:
            with open(path, "a", encoding="utf-8") as f:
                f.write("[")
                records = []
                for rec in self.from_xml():
                    records.append(rec)
                    if len(records) > 1000:
                        f.write(json.dumps(records, **params)[1:-1] + sep)
                        records = []
                if records:
                    f.write(json.dumps(records, **params)[1:-1])
                f.write("]")
        except KeyboardInterrupt as exc:
            # Remove the partial JSON file if write is interrupted
            os.remove(path)
            raise IOError("Conversion to JSON failed") from exc

    def counts(self):
        """Counts the number of records in each file"""
        counts = {}
        for filelike in self.files:
            with open(filelike.path, mode="r", encoding="utf8") as f:
                with mmap.mmap(f.fileno(), length=0, access=mmap.ACCESS_READ) as m:
                    counts[filelike.path] = len(re.findall(rb"\n  <tuple>", m.read()))
        return counts[list(counts)[0]] if len(counts) == 1 else counts

    def verify_group(self, path, module=None):
        """Verifies that all fields in a group are present in the export

        Raises
        ------
        ValueError
            if one or more fields missing
        """
        if module is None:
            module = self.module
        group = tuple(self.schema.get_field_info(module, path).get("GroupFields", []))
        missing = set(group) - set(self.fields)
        if missing:
            raise ValueError(f"Group including '{path}' is missing fields: {missing}")

    def report_progress(self, by="time", at=5):
        """Prints progress notification messages when reading a file

        Parameters
        ----------
        by : str
            either "count" or "time"
        at : int
            number of seconds (if by time) or number of records (if by count)
        """
        if self._notify_start is None:
            self._job_start = time.time()
            self._notify_start = time.time()

        elapsed = time.time() - (
            self._job_start if self._job_done else self._notify_start
        )
        if (
            self._job_done
            or by == "time"
            and elapsed >= at
            or by == "count"
            and self._notify_count
            and not (self._notify_count % at)
        ):
            print(
                "{:,} records processed (t{}={:.1f}s)".format(
                    self._notify_count, "otal" if self._job_done else "", elapsed
                )
            )
            self._notify_start = time.time()

    def _parse(self, xml):
        """Parses a record from XML

        Parameters
        ----------
        xml : lxml.Element
            XML representing a single record

        Returns
        -------
        dict
           EMu record
        """
        if self._rec_class != dict:
            dct = self._rec_class(module=self.module)
        else:
            dct = self._rec_class()

        elements = [(dct, "", xml)]
        while elements:
            new_elems = []
            for obj, parent_name, elem in elements:
                for child in elem:
                    # Add an empty rows to a nested table, which do not contain
                    # child nodes when exported from EMu
                    if child is None:
                        obj.append(None)
                        continue

                    # Get field name
                    name = child.get("name")
                    if name is None:
                        name = ""

                    # Get field text
                    text = child.text
                    if text is not None:
                        text = text.strip()

                    # Add an atomic field
                    if child.tag == "atom":
                        try:
                            obj[name] = text
                        except TypeError:
                            obj.append(text)

                    # Add a reference
                    elif child.tag == "tuple" and is_ref(name) and not is_tab(name):
                        obj[name] = {}
                        new_elems.append((obj[name], name, child))

                    # Add a table or reference table
                    elif child.tag == "table" or (child.tag == "tuple" and name):
                        try:
                            obj[name] = []
                            new_elems.append((obj[name], name, child))
                        except TypeError:
                            obj.append([])
                            new_elems.append((obj[-1], name, child))

                    # Add a row to a reference table
                    elif (
                        child.tag == "tuple"
                        and is_ref_tab(parent_name)
                        and not is_nesttab(parent_name)
                    ):
                        obj.append({})
                        new_elems.append((obj[-1], name, child))

                    # Add an empty row to an outer nested table
                    elif (
                        child.tag == "tuple"
                        and is_nesttab(parent_name)
                        and not len(child)
                    ):
                        new_elems.append((obj, name, [None]))

                    elif child.tag == "tuple":
                        new_elems.append((obj, name, child))

                elements = new_elems

        return dct

    def _get_files(self):
        """Analyzes source files on self.path"""
        files = []
        zip_file = None
        if self.path:
            if os.path.isdir(self.path):
                files = glob.glob(os.path.join(self.path, "*.xml"))
            elif self.path.lower().endswith(".xml"):
                files = [self.path]
            elif self.path.lower().endswith(".zip"):
                zip_file = zipfile.ZipFile(self.path)
                files = zipfile.ZipFile(self.path).infolist()
            else:
                raise IOError(f"Invalid path: {self.path}")

        # Order source files from oldest to newest
        self.files = [FileLike(obj, zip_file=zip_file) for obj in files]
        self.files.sort(key=lambda f: f.getmtime())

        # Get the module name from the first table tag
        with self.files[0].open(encoding="utf-8") as f:
            for line in f:
                if line.strip().startswith("<table"):
                    self.module = line.split("=", 1)[-1].strip('">\r\n')
                    break

    def _load_schema(self):
        """Tries to load the schema based on the rec_class"""
        if self.schema is None:
            try:
                schema = self._rec_class.schema
                if schema is None:
                    # This will also load the configuration
                    schema = self._rec_class(module=self.module).schema
            except (AttributeError, ValueError):
                schema = None
        return self.schema

    def _parse_file_schema(self):
        """Parses top-level fields from header of EMu XML file

        Returns
        -------
        list
            list of top-level fields
        """
        fields = {}
        for filelike in self.files:
            with open(filelike.path, "r", encoding="utf-8") as f:
                lines = []
                for line in f:
                    lines.append(line)
                    if line.startswith("?>"):
                        break
                content = "".join(lines)

            tables = []
            for line in (
                re.search(r"<?schema\s+(.*?)\?>", content, flags=re.DOTALL)
                .group(1)
                .splitlines()
            ):
                line = line.strip()
                try:
                    dtype, field = [s.strip() for s in line.rsplit(" ", 1)]
                except ValueError:
                    dtype = None
                    tables.pop()
                else:
                    if dtype == "table":
                        tables.append(field)
                    else:
                        segments = tables[1:] + [field]
                        segments = [
                            s
                            for i, s in enumerate(segments)
                            if strip_tab(s) not in {strip_tab(s) for s in segments[:i]}
                        ]
                        fields[segments[0]] = 1
        return tuple(fields)


class FileLike:
    """Open text and zip files using the same interface

    Parameters
    ----------
    filelike : mixed
        either the path to an XML file or a ZipInfo object
    zip_file : zipfile.ZipFile
        if filelike is a ZipInfo object, the zip file containing that object

    Attributes
    ----------
    path : str
        path to file
    zip_info : zipfile.ZipInfo
        member of a zip archive
    zip_file : zipfile.ZipFile
        the zip file containing the ZipInfo object
    """

    def __init__(self, filelike, zip_file=None):
        self.path = None
        self.zip_info = None
        self.zip_file = None
        if zip_file:
            self.zip_info = filelike
            self.zip_file = zip_file
        else:
            self.path = os.path.realpath(filelike)

    def __str__(self):
        return f'<FileLike name="{self.filename}">'

    def __repr__(self):
        return str(self)

    @property
    def filename(self):
        """Name of the file-like object"""
        return os.path.basename(self.path) if self.path else self.zip_info.filename

    def open(self, mode="r", encoding=None):
        """Opens a file or ZipInfo object"""
        if not self.zip_info:
            return open(self.path, mode=mode, encoding=encoding)
        stream = self.zip_file.open(self.zip_info, mode.rstrip("b"))
        if encoding:
            return _ByteDecoder(stream, encoding)
        return stream

    def getmtime(self):
        """Returns last modification timestamp from a file or ZipInfo object"""
        try:
            return os.path.getmtime(self.path)
        except TypeError:
            return dt.datetime(*self.zip_info.date_time).timestamp()


class _ByteDecoder:
    """File-like context manager that encodes a binary stream from a zip file"""

    def __init__(self, stream, encoding):
        self._stream = stream
        self._encoding = encoding

    def __iter__(self):
        for line in self._stream:
            yield line.decode(self._encoding)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exception, traceback):
        if exception:
            raise exception
        self._stream.close()


def write_csv(records, path, **kwargs):
    """Writes records to CSV

    Parameters
    ----------
    records : list-like
        list of EMuRecords to be written
    path : str
        path to write the CSV file
    kwargs :
        any keyword argument accepted by open()
    """
    flattened = [flatten(r) for r in records]

    keys = {}
    for rec in flattened:
        keys.update({k: 1 for k in rec})

    # Reorder keys to account for varying grid lengths
    grouped = {}
    for key in keys:
        grouped.setdefault(re.sub(r"\.\d+\.", ".x.", key), []).append(key)

    grouped = {
        k: sorted(v, key=lambda s: ".".join([s.zfill(8) for s in s.split(".")]))
        for k, v in grouped.items()
    }

    fieldnames = []
    for group in grouped.values():
        fieldnames.extend(group)

    kwargs.setdefault("encoding", "utf-8-sig")
    kwargs.setdefault("newline", "")
    with open(path, "w", **kwargs) as f:
        writer = csv.DictWriter(f, fieldnames=fieldnames)
        writer.writeheader()
        writer.writerows(({k: r.get(k) for k in fieldnames} for r in flattened))


def write_import(*args, **kwargs):
    """Writes records to an EMu import file

    Alias for write_xml()
    """
    return write_xml(*args, **kwargs)


def write_xml(records, path, **kwargs):
    """Writes records to an EMu import file

    Parameters
    ----------
    records : list-like
        list of EMuRecords to be imported
    path : str
        path to write the import file
    kwargs :
        any keyword argument accepted by the to_xml() method of the record class
    """
    root = etree.Element("table")
    root.set("name", records[0].module)
    root.addprevious(etree.Comment(" Data "))

    for i, rec in enumerate(records):
        try:
            node = rec.copy().to_xml(root, **kwargs)
            node.addprevious(etree.Comment(f" Row {i + 1} "))
        except AttributeError:
            raise ValueError(f"Could not convert record to XML: {rec}")

    root.getroottree().write(
        path, pretty_print=True, xml_declaration=True, encoding="utf-8"
    )


def write_group(records, path, irn=None, name=None):
    """Writes an import for the egroups module

    Parameters
    ----------
    records : list
        list of EMuRecords, each of which specifies an irn
    path : str
        path to write the import file
    irn : int
        the irn of an existing egroups record (updates only)
    name : str
        the name of the group
    """
    if not irn and not name:
        raise ValueError("Must specify at least one of irn or name for a group")
    rec = records[0].__class__(
        {
            "GroupType": "Static",
            "Module": records[0].module,
            "Keys_tab": [rec["irn"] for rec in records],
        },
        module="egroups",
    )
    if irn:
        rec["irn"] = irn
    if name:
        rec["GroupName"] = name
    write_import([rec], path)



import json
import os
import pickle
import re
import zipfile
from datetime import date, datetime, time, timedelta

import pytest
from lxml import etree

from xmu import (
    EMuColumn,
    EMuConfig,
    EMuDate,
    EMuEncoder,
    EMuFloat,
    EMuLatitude,
    EMuLongitude,
    EMuReader,
    EMuRecord,
    EMuRow,
    EMuSchema,
    EMuTime,
    get_mod,
    has_mod,
    is_nesttab,
    is_nesttab_inner,
    is_ref,
    is_ref_tab,
    is_tab,
    strip_mod,
    strip_tab,
    write_xml,
    write_group,
)
from xmu.types import ExtendedDate

os.chdir("tests")


@pytest.fixture(scope="session")
def output_dir(tmp_path_factory):
    return tmp_path_factory.mktemp("output")


@pytest.fixture
def schema_file(output_dir):
    # This is a partial schema that omits keys not used by the application
    pl = """#
#
#

#
#
#
#
use utf8;

%Schema =
(
	emain =>
	{
		table => 'emain',
		columns =>
		{
            'EmuClientTable_tab' =>
			{
				ColumnName => 'EmuClientTable_tab',
				DataType => 'Text',
                RefLink => 'EmuClientTableRef_tab'

				ItemName => 'Client Table',
			},
            'EmuClientTableRef_tab' =>
			{
				ColumnName => 'EmuRef_tab',
				DataType => 'Integer',
                RefLink => 'eref'

				ItemName => 'Client Reference Table',
			},
			'EmuDate0' =>
			{
				ColumnName => 'EmuDate0',
				DataType => 'Date',

                ItemName => 'Date',
				ItemFields =>
				[
					[ 8, 2, 2 ],
					[ 8, 2, 2 ],
					[ 8, 2, 2 ],
				],
			},
            'EmuEmpty' =>
			{
				ColumnName => 'EmuEmpty',
				DataType => 'Text',

                ItemName => 'Empty',
				ItemCount => 1,
				ItemFields => [ 15 ],
			},
            'EmuInteger' =>
			{
				ColumnName => 'EmuInteger',
				DataType => 'Integer',

                ItemName => 'Integer',
			},
			'EmuFloat' =>
			{
				ColumnName => 'EmuFloat',
				DataType => 'Float',

                ItemName => 'Float',
			},
            'EmuLatitude' =>
			{
				ColumnName => 'EmuLatitude',
				DataType => 'Latitude',

				ItemName => 'Latitude',
			},
            'EmuLongitude' =>
			{
				ColumnName => 'EmuLongitude',
				DataType => 'Longitude',

				ItemName => 'Longitude',
			},
            'EmuLookupParent' =>
			{
				ColumnName => 'EmuText',
				DataType => 'Text',
                LookupName => 'Lookup',
				LookupParent => 'SecLookupRoot',

				ItemName => 'Lookup Parent',
			},
            'EmuLookupChild' =>
			{
				ColumnName => 'EmuText',
				DataType => 'Text',
                LookupName => 'Lookup',
				LookupParent => 'EmuLookupParent',

				ItemName => 'Lookup Child',
			},
            'EmuLookupGrandchild' =>
			{
				ColumnName => 'EmuText',
				DataType => 'Text',
                LookupName => 'Lookup',
				LookupParent => 'EmuLookupChild',

				ItemName => 'Lookup Grandchild',
			},
            'EmuNestedTable_nesttab' =>
			{
				ColumnName => 'EmuNestedTable_nesttab',
				DataType => 'Text',

                ItemName => 'Nested Table',
			},
            'EmuNotVisible' =>
			{
				ColumnName => 'EmuNotVisible',
				DataType => 'Text',
			},
            'EmuRef' =>
			{
				ColumnName => 'EmuEmpty',
				DataType => 'Integer',
                RefTable => 'eref',

                ItemName => 'Reference',
				ItemFields =>
				[
					  10,   10,   10,   10,   10,
					  10,   10,   10,   10,   10,
                      10,
				],
			},
            'EmuRef_nesttab' =>
			{
				ColumnName => 'EmuRef_nesttab',
				DataType => 'Integer',
                RefTable => 'eref'

				ItemName => 'Nested Reference Table',
			},
            'EmuRef_tab' =>
			{
				ColumnName => 'EmuRef_tab',
				DataType => 'Integer',
                RefTable => 'eref'

				ItemName => 'Reference Table',
			},
			'EmuTable_tab' =>
			{
				ColumnName => 'EmuTable_tab',
				DataType => 'Text',

                ItemName => 'Table',
			},
            'EmuTableUngrouped_tab' =>
			{
				ColumnName => 'EmuTable_tab',
				DataType => 'Text',

                ItemName => 'Table Ungrouped',
			},
            'EmuText' =>
			{
				ColumnName => 'EmuText',
				DataType => 'Text',

				ItemName => 'Text',
			},
            'EmuTime0' =>
			{
				ColumnName => 'EmuTime0',
				DataType => 'Time',

                ItemName => 'Time',
				ItemFields =>
				[
					[ 8, 2, 2 ],
					[ 8, 2, 2 ],
					[ 8, 2, 2 ],
				],
			},
			'irn' =>
			{
				ColumnName => 'irn',
				DataType => 'Integer',

				ItemName => 'IRN',
			},
		},
        groups =>
        {
            'EmuGrid_grp' =>
            [
                'EmuDate0',
                'EmuNestedTable_nesttab',
                'EmuTable_tab',
                'EmuRef_tab'
            ],
            'EmuClientGrid_grp' =>
            [
                'EmuClientTable_tab',
            ],
        },
	},
	eref =>
	{
		table => 'eref',
		columns =>
		{
            'EmuRefOnly' =>
			{
				ColumnName => 'EmuRefOnly',
				DataType => 'Text',

				ItemName => 'Reference only',
			},
            'EmuTableInRef_tab' =>
			{
				ColumnName => 'EmuTableInRef',
				DataType => 'Text',

				ItemName => 'Table in Reference',
			},
            'irn' =>
			{
				ColumnName => 'irn',
				DataType => 'Integer',

				ItemName => 'IRN',
			},
		},
        groups =>
        {
            'EmuGridInReference_grp' =>
            [
                'EmuTableInRef_tab'
            ],
        },
	},
);

1;
"""
    path = output_dir / "schema.pl"
    with open(path, "w") as f:
        f.write(pl)
    return str(path)


@pytest.fixture
def config_file(schema_file):
    config = EMuConfig(".")
    config["schema_path"] = schema_file
    config["groups"]["emain"] = {
        "EmuGrid_tab": [
            "EmuDate0",
            "EmuNestedTable_nesttab",
            "EmuTable_tab",
            "EmuRef_tab",
        ]
    }
    config["lookup_no_autopopulate"] = ["emain.EmuLookupParent", "emain.EmuLookupChild"]
    config.save_rcfile(".", overwrite=True)
    return ".xmurc"


@pytest.fixture
def xml_file(output_dir):
    xml = """<?xml version="1.0" encoding="UTF-8" ?>
<?schema
  table           emain
    integer         irn
    text short      EmuText
    integer         EmuInteger
    float           EmuFloat
    latitude        EmuLatitude
    longitude       EmuLongitude
    table           EmuRef
      integer         irn
      text short      EmuRefOnly
    end
    table           EmuDate0
      date            EmuDate
    end
    table           EmuTime0
      time            EmuTime
    end
    table           EmuTable_tab
      text short      EmuTable
    end
    table           EmuTableUngrouped_tab
      text short      EmuTableUngrouped
    end
    table           EmuRef_tab
      table           EmuRef
        integer         irn
        text short      EmuRefOnly
      end
    end
    table           EmuNestedTable_nesttab
      table           EmuNestedTable_nesttab_inner
        text short      EmuNestedTable
      end
    end
  end
?>
<table name="emain">

  <!-- Row 1 -->
  <tuple>
    <atom name="irn">1000000</atom>
    <atom name="EmuText">Text</atom>
    <atom name="EmuInteger">1</atom>
    <atom name="EmuFloat">1.0</atom>
    <atom name="EmuLatitude">45 30 15 N</atom>
    <atom name="EmuLongitude">-130 10 5 W</atom>
    <tuple name="EmuRef">
      <atom name="irn">1000000</atom>
      <atom name="EmuRefOnly">Text</atom>
    </tuple>
    <table name="EmuDate0">
      <tuple>
        <atom name="EmuDate">1970-01-01</atom>
      </tuple>
      <tuple>
        <atom name="EmuDate">Jan 1970</atom>
      </tuple>
      <tuple>
        <atom name="EmuDate">1970</atom>
      </tuple>
    </table>
    <table name="EmuTime0">
      <tuple>
        <atom name="EmuTime">9:00</atom>
      </tuple>
      <tuple>
        <atom name="EmuTime">12:00</atom>
      </tuple>
      <tuple>
        <atom name="EmuTime">15:00</atom>
      </tuple>
    </table>
    <table name="EmuTable_tab">
      <tuple>
        <atom name="EmuTable">Text</atom>
      </tuple>
      <tuple>
        <atom name="EmuTable">Text</atom>
      </tuple>
      <tuple>
      </tuple>
    </table>
    <table name="EmuTableUngrouped_tab">
      <tuple>
        <atom name="EmuTableUngrouped">Text</atom>
      </tuple>
    </table>
    <table name="EmuRef_tab">
      <tuple>
      </tuple>
      <tuple>
      </tuple>
      <tuple>
        <atom name="irn">1000000</atom>
        <atom name="EmuRefOnly">Text</atom>
      </tuple>
    </table>
    <table name="EmuNestedTable_nesttab">
      <tuple>
      </tuple>
      <tuple>
        <table name="EmuNestedTable_nesttab_inner">
          <tuple>
            <atom name="EmuNestedTable">Text</atom>
          </tuple>
        </table>
      </tuple>
    </table>
  </tuple>
</table>
"""
    path = output_dir / "xmldata.xml"
    with open(path, "w") as f:
        f.write(xml)
    return str(path)


@pytest.fixture
def rec(xml_file):
    reader = EMuReader(xml_file)
    for rec in reader:
        return EMuRecord(rec, module=reader.module)


@pytest.fixture
def expected_rec():
    # Expected when using rec_class == dict
    return {
        "irn": "1000000",
        "EmuText": "Text",
        "EmuInteger": "1",
        "EmuFloat": "1.0",
        "EmuLatitude": "45 30 15 N",
        "EmuLongitude": "-130 10 5 W",
        "EmuRef": {"irn": "1000000", "EmuRefOnly": "Text"},
        "EmuDate0": ["1970-01-01", "Jan 1970", "1970"],
        "EmuTime0": ["9:00", "12:00", "15:00"],
        "EmuTable_tab": ["Text", "Text"],
        "EmuTableUngrouped_tab": ["Text"],
        "EmuRef_tab": [{}, {}, {"irn": "1000000", "EmuRefOnly": "Text"}],
        "EmuNestedTable_nesttab": [None, ["Text"]],
    }


@pytest.fixture
def grid(rec):
    return rec.grid("EmuTable_tab").pad()


def test_config(config_file, output_dir):
    config = EMuConfig(config_file)
    assert len(config) == 4
    assert [k for k in config] == [
        "schema_path",
        "groups",
        "make_visible",
        "lookup_no_autopopulate",
    ]
    assert config["schema_path"] == str(output_dir / "schema.pl")
    assert config["make_visible"] == []
    del config["make_visible"]
    assert "make_visible" not in config


def test_schema(schema_file):
    schema_pl = EMuSchema(schema_file)
    json_path = os.path.splitext(schema_file)[0] + ".json"
    schema_pl.to_json(json_path)
    assert schema_pl == EMuSchema(json_path)


def test_schema_from_config(config_file, schema_file):
    EMuConfig(config_file)
    assert EMuSchema() == EMuSchema(schema_file)


@pytest.mark.skip(reason="picks up global config file")
def test_schema_no_args():
    EMuSchema.config = None
    assert EMuSchema() == {}


@pytest.mark.skip(reason="picks up global config file")
def test_schema_from_args():
    schema = {"Schema": {}}
    assert EMuSchema(schema) == schema


def test_schema_get(schema_file):
    schema = EMuSchema(schema_file)
    assert schema.get("Schema.emain.columns.EmuDate0") == {
        "ColumnName": "EmuDate0",
        "DataType": "Date",
        "ItemName": "Date",
        "ItemFields": [[8, 2, 2], [8, 2, 2], [8, 2, 2]],
        "GroupFields": [
            "EmuDate0",
            "EmuNestedTable_nesttab",
            "EmuTable_tab",
            "EmuRef_tab",
        ],
    }
    assert schema.get("Schema.emain.columns.EmuInvalid") is None


def test_schema_iterfields(schema_file):
    schema = EMuSchema(schema_file)
    fields = {}
    for module, field, info in schema.iterfields():
        fields[(module, field)] = info
    assert list(fields) == [
        ("emain", "EmuClientTable_tab"),
        ("emain", "EmuClientTableRef_tab"),
        ("emain", "EmuDate0"),
        ("emain", "EmuEmpty"),
        ("emain", "EmuInteger"),
        ("emain", "EmuFloat"),
        ("emain", "EmuLatitude"),
        ("emain", "EmuLongitude"),
        ("emain", "EmuLookupParent"),
        ("emain", "EmuLookupChild"),
        ("emain", "EmuLookupGrandchild"),
        ("emain", "EmuNestedTable_nesttab"),
        ("emain", "EmuNotVisible"),
        ("emain", "EmuRef"),
        ("emain", "EmuRef_nesttab"),
        ("emain", "EmuRef_tab"),
        ("emain", "EmuTable_tab"),
        ("emain", "EmuTableUngrouped_tab"),
        ("emain", "EmuText"),
        ("emain", "EmuTime0"),
        ("emain", "irn"),
        ("eref", "EmuRefOnly"),
        ("eref", "EmuTableInRef_tab"),
        ("eref", "irn"),
    ]


def test_schema_getitem_bad_module(schema_file):
    match = (
        r"Path not found: \('Schema', 'einvalid', 'columns', 'EmuText'\)"
        r" \(failed at einvalid\)"
    )
    with pytest.raises(KeyError, match=match):
        EMuSchema(schema_file).get_field_info("einvalid", "EmuText")


def test_schema_getitem_not_visible(schema_file):
    with pytest.raises(KeyError, match=r"emain.EmuNotVisible is valid but"):
        EMuSchema(schema_file).get_field_info("emain", "EmuNotVisible")


def test_col_change():
    col = EMuColumn(["Text"], module="emain", field="EmuTable_tab")

    col = col + ["Text"]
    assert col == ["Text"] * 2

    col.insert(0, "Text")
    assert col == ["Text"] * 3

    col.append("Text")
    assert col == ["Text"] * 4

    col.extend(["Text"])
    assert col == ["Text"] * 5

    col += ["Text"]
    assert col == ["Text"] * 6

    assert isinstance(col, EMuColumn)


def test_col_to_xml():
    col = EMuColumn(["Text"], module="emain", field="EmuTable_tab")
    assert (
        etree.tostring(col.to_xml())
        == b'<table name="EmuTable_tab"><tuple row="1"><atom name="EmuTable">Text</atom></tuple></table>'
    )


def test_col_no_module():
    with pytest.raises(ValueError, match=r"Must provide module when schema is used"):
        EMuColumn()


def test_row(rec):
    row = rec.grid("EmuTable_tab")[0]
    assert row["EmuDate0"] == rec["EmuDate0"][0] == EMuDate("1970-01-01")
    del row["EmuDate0"]
    assert row["EmuDate0"] is None and rec["EmuDate0"][0] is None


def test_row_id():
    rec = EMuRecord(
        {
            "irn": 1000000,
            "EmuTable_tab(+)": ["Text", "Text", "Text"],
            "EmuRef_tab(+)": [
                {"irn": 1000000},
                {"irn": 1000000},
                {"irn": 1000001},
            ],
        },
        module="emain",
    )
    row_ids = []
    for i in range(2):
        row_ids.append([r.row_id() for r in rec.grid("EmuTable_tab")])
    assert row_ids[0] == row_ids[1]
    assert row_ids[0][0] != row_ids[0][1]
    assert row_ids[0][0] != row_ids[0][2]
    assert row_ids[0][1] != row_ids[0][2]


def test_row_in_reference(rec):
    rec = rec.copy()
    row1 = EMuRow(rec, "EmuRef.EmuTableInRef_tab", 0)
    row2 = EMuRow(rec["EmuRef"], "EmuTableInRef_tab", 0)
    assert row1 == row2


def test_row_from_ungrouped(rec):
    with pytest.raises(KeyError, match=r"'emain.EmuText is not part of a group"):
        EMuRow(rec, "EmuText", 0)


def test_grid_by_index(grid):
    assert grid[0] == {
        "EmuDate0": EMuDate("1970-01-01"),
        "EmuNestedTable_nesttab": [],
        "EmuRef_tab": {},
        "EmuTable_tab": "Text",
    }
    assert grid[1] == {
        "EmuDate0": EMuDate("Jan 1970"),
        "EmuNestedTable_nesttab": ["Text"],
        "EmuRef_tab": {},
        "EmuTable_tab": "Text",
    }
    assert grid[2] == {
        "EmuDate0": EMuDate("1970"),
        "EmuNestedTable_nesttab": [],
        "EmuRef_tab": {"irn": 1000000, "EmuRefOnly": "Text"},
        "EmuTable_tab": "",
    }


def test_grid_by_str(grid):
    assert grid["EmuDate0"] == [
        EMuDate("1970-01-01"),
        EMuDate("Jan 1970"),
        EMuDate("1970"),
    ]


def test_empty_grid(rec):
    rec = rec.copy()

    rec["EmuRef"] = {"EmuTableInRef_tab": []}
    grid = rec.grid("EmuRef.EmuTableInRef_tab")
    assert len(grid) == 0

    rec["EmuRef"] = {}
    grid = rec.grid("EmuRef.EmuTableInRef_tab")
    assert len(grid) == 0


def test_grid_query(grid):
    results = grid.query("EmuRef_tab", where={"EmuTable_tab": "Text"})
    assert results == [{}, {}]


def test_grid_get_item(rec):
    grid = rec.grid("EmuTable_tab").pad()
    assert len(grid[{"EmuDate0": "1970-01-01", "EmuTable_tab": "Text"}]) == 1
    assert (
        len(grid[{"EmuDate0": ["1970-01-01", "Jan 1970"], "EmuTable_tab": "Text"}]) == 2
    )
    assert (
        len(grid[{"EmuDate0": ["1970-01-01", "Jan 1970"], "EmuTable_tab": "Number"}])
        == 0
    )


def test_grid_del_item(rec):
    grid = rec.grid("EmuTable_tab").pad()
    del grid[0]
    assert rec["EmuDate0"] == [EMuDate("Jan 1970"), EMuDate("1970")]
    assert rec["EmuNestedTable_nesttab"] == [["Text"], []]
    assert rec["EmuRef_tab"] == [{}, {"irn": 1000000, "EmuRefOnly": "Text"}]
    assert rec["EmuTable_tab"] == ["Text", ""]


def test_grid_from_client_table(rec):
    assert "GroupFields" in rec.schema.get_field_info(
        rec.module, "EmuClientTableRef_tab"
    )
    assert "GroupFields" not in rec.schema.get_field_info(
        rec.module, "EmuClientTable_tab"
    )


def test_grid_in_reference(rec):
    rec = rec.copy()
    rec["EmuRef"] = {"EmuTableInRef_tab": ["Text", "Text"]}
    grid1 = rec.grid("EmuRef.EmuTableInRef_tab")
    grid2 = rec["EmuRef"].grid("EmuTableInRef_tab")
    assert grid1
    assert grid1 == grid2


def test_grid_inconsistent_modifier():
    rec = EMuRecord(
        {"EmuDate0(+)": ["1970-01-01"], "EmuRef_tab": [1234567]}, module="emain"
    )
    with pytest.raises(ValueError, match="Inconsistent modifier within grid"):
        rec.grid("EmuRef_tab").columns


def test_grid_multiple_modifiers():
    rec = EMuRecord(
        {"EmuDate0(+)": ["1970-01-01"], "EmuRef_tab(-)": [1234567]}, module="emain"
    )
    with pytest.raises(ValueError, match="Inconsistent modifier within grid"):
        rec.grid("EmuRef_tab").columns


def test_grid_set_item(grid):
    with pytest.raises(NotImplementedError, match="Cannot set items on an EMuGrid"):
        grid[0] = None


def test_grid_insert(grid):
    with pytest.raises(NotImplementedError, match="Cannot insert into an EMuGrid"):
        grid.insert(0, None)


def test_grid_from_ungrouped(rec):
    with pytest.raises(KeyError, match=r"'emain.EmuText is not part of a group"):
        rec.grid("EmuText")


def test_grid_missing_fields(xml_file):
    reader = EMuReader(xml_file)
    reader.verify_group("EmuTable_tab")
    for rec in reader:
        rec = EMuRecord(rec, module=reader.module)
        grid = rec.grid("EmuTable_tab")
        del rec["EmuTable_tab"]
        with pytest.raises(ValueError, match="Grid including"):
            grid.verify()


@pytest.mark.parametrize(
    "key,val,expected",
    [
        (
            "EmuLookupParent",
            "Text",
            '<tuple><atom name="EmuLookupParent">Text</atom></tuple>',
        ),
        (
            "EmuLookupChild",
            "Text",
            r'<tuple><atom name="EmuLookupChild">Text</atom><atom name="EmuLookupParent"></atom></tuple>',
        ),
        (
            "EmuLookupGrandchild",
            "Text",
            r'<tuple><atom name="EmuLookupGrandchild">Text</atom><atom name="EmuLookupChild"></atom><atom name="EmuLookupParent"></atom></tuple>',
        ),
    ],
)
def test_lookup(key, val, expected):
    rec = EMuRecord({key: val}, module="emain")
    assert etree.tostring(rec.to_xml()).decode("utf-8") == expected


def test_parse_file_schema(output_dir):
    reader = EMuReader(output_dir)
    assert reader.fields == (
        "irn",
        "EmuText",
        "EmuInteger",
        "EmuFloat",
        "EmuLatitude",
        "EmuLongitude",
        "EmuRef",
        "EmuDate0",
        "EmuTime0",
        "EmuTable_tab",
        "EmuTableUngrouped_tab",
        "EmuRef_tab",
        "EmuNestedTable_nesttab",
    )


def test_read_from_dir(xml_file, output_dir, expected_rec):
    reader = EMuReader(output_dir)
    assert [str(f) for f in reader.files] == ['<FileLike name="xmldata.xml">']
    for rec in reader:
        assert rec == expected_rec


def test_read_from_parallel(xml_file, expected_rec):
    def callback(path):
        for rec in EMuReader(path):
            return rec["irn"]

    results = EMuReader(xml_file).from_xml_parallel(callback, num_parts=8)
    assert results == ["1000000"]


def test_read_from_parallel_list(output_dir):
    def callback(path):
        results = []
        for rec in EMuReader(path):
            results.append(EMuRecord(rec, module="emain"))
        return results

    xml_path = str(output_dir / "xmldata_parallel_list.xml")
    records = []
    for i in range(0, 64):
        records.append(EMuRecord({"irn": i}, module="emain"))
    write_xml(records, xml_path, kind="emu")
    results = EMuReader(xml_path).from_xml_parallel(callback, num_parts=8)
    for i, rec in enumerate(results):
        assert rec["irn"] == i


def test_read_from_parallel_dict_combine(output_dir):
    def callback(path):
        results = {}
        for i, _ in enumerate(EMuReader(path)):
            results[i] = 1
        return results

    xml_path = str(output_dir / "xmldata_parallel_dict_combine.xml")
    records = []
    for i in range(0, 64):
        records.append(EMuRecord({"irn": i}, module="emain"))
    write_xml(records, xml_path, kind="emu")
    results = EMuReader(xml_path).from_xml_parallel(
        callback, num_parts=8, handle_repeated_keys="combine"
    )
    assert results[0] == [1] * 8


def test_read_from_parallel_dict_keep(output_dir):
    def callback(path):
        results = {}
        for rec in EMuReader(path):
            results["irn"] = rec["irn"]
            break
        return results

    xml_path = str(output_dir / "xmldata_parallel_dict_keep.xml")
    records = []
    for i in range(0, 64):
        records.append(EMuRecord({"irn": i}, module="emain"))
    write_xml(records, xml_path, kind="emu")
    results = EMuReader(xml_path).from_xml_parallel(
        callback, num_parts=8, handle_repeated_keys="keep"
    )
    assert results["irn"] == "0"


def test_read_from_parallel_dict_overwrite(output_dir):
    def callback(path):
        results = {}
        for rec in EMuReader(path):
            results["irn"] = rec["irn"]
        return results

    xml_path = str(output_dir / "xmldata_parallel_dict_overwite.xml")
    records = []
    for i in range(0, 64):
        records.append(EMuRecord({"irn": i}, module="emain"))
    write_xml(records, xml_path, kind="emu")
    results = EMuReader(xml_path).from_xml_parallel(
        callback, num_parts=8, handle_repeated_keys="overwrite"
    )
    assert results["irn"] == "63"


def test_read_from_parallel_dict_raise(output_dir):
    def callback(path):
        results = {}
        for i, _ in enumerate(EMuReader(path)):
            results[i] = 1
        return results

    xml_path = str(output_dir / "xmldata_parallel_dict_combine.xml")
    records = []
    for i in range(0, 64):
        records.append(EMuRecord({"irn": i}, module="emain"))
    write_xml(records, xml_path, kind="emu")
    with pytest.raises(KeyError, match=r"Duplicate keys returned"):
        EMuReader(xml_path).from_xml_parallel(
            callback, num_parts=8, handle_repeated_keys="raise"
        )


def test_read_from_json(xml_file, output_dir):
    reader = EMuReader(xml_file)
    for rec in reader:
        rec_from_xml = EMuRecord(rec, module=reader.module)

    simple_rec = EMuRecord({"irn": 1234567}, module="emain")
    records = [rec_from_xml] + [simple_rec] * 5000
    write_xml(records, output_dir / "xmldata_5000.xml", kind="emu")

    xml_path = str(output_dir / "xmldata_5000.xml")
    json_path = str(output_dir / "xmldata_5000.json")

    reader = EMuReader(xml_path, json_path=json_path)
    for rec in reader:
        rec_from_json = EMuRecord(rec, module=reader.module)
        break

    rec_from_json_chunked = None
    for rec in reader.from_json(chunk_size=8192):
        if not rec_from_json_chunked:
            rec_from_json_chunked = EMuRecord(rec, module=reader.module)

    assert rec_from_json == rec_from_xml
    assert rec_from_json_chunked == rec_from_xml


def test_read_from_zip(xml_file, output_dir, expected_rec):
    path = str(output_dir / "xmldata.zip")
    with zipfile.ZipFile(path, "w") as f:
        f.write(xml_file, arcname="xmldata_1.xml")
        f.write(xml_file, arcname="xmldata_2.xml")
    reader = EMuReader(path)
    assert [repr(f) for f in reader.files] == [
        '<FileLike name="xmldata_1.xml">',
        '<FileLike name="xmldata_2.xml">',
    ]
    for rec in reader:
        assert rec == expected_rec


def test_rec_round_trip(rec, output_dir):
    path = str(output_dir / "import.xml")
    write_xml([rec], path, kind="emu")
    reader = EMuReader(path)
    for rec_ in reader:
        assert EMuRecord(rec_, module=reader.module) == rec


def test_len(output_dir):
    records = [EMuRecord({"irn": 1234567}, module="emain")] * 100
    path = output_dir / "xmldata_len.xml"
    write_xml(records, path, kind="emu")
    assert len(EMuReader(path)) == 100


@pytest.mark.parametrize(
    "path",
    [
        ("EmuRef", "EmuRefOnly"),
        ["EmuRef", "EmuRefOnly"],
        "EmuRef.EmuRefOnly",
        "EmuRef/EmuRefOnly",
    ],
)
def test_rec_getitem(rec, path):
    assert rec[path] == "Text"


@pytest.mark.parametrize(
    "path",
    [None, {}],
)
def test_rec_getitem_invalid(rec, path):
    with pytest.raises(ValueError, match=r"Invalid path format:"):
        rec[path]


def test_report_progress(xml_file, output_dir, capsys):
    reader = EMuReader(output_dir)
    for i, _ in enumerate(reader):
        reader.report_progress("count", 10000)
    assert re.match(r"\d+(,\d{3})* records processed \(total=", capsys.readouterr().out)


def test_write_csv(xml_file, output_dir):
    reader = EMuReader(xml_file)
    path = str(output_dir / "test.csv")
    reader.to_csv(path)
    with open(path, encoding="utf-8-sig", newline="") as f:
        assert f.read().splitlines() == [
            "irn,EmuText,EmuInteger,EmuFloat,EmuLatitude,EmuLongitude,EmuRef.irn,EmuRef.EmuRefOnly,EmuDate0.1.EmuDate,EmuDate0.2.EmuDate,EmuDate0.3.EmuDate,EmuTime0.1.EmuTime,EmuTime0.2.EmuTime,EmuTime0.3.EmuTime,EmuTable_tab.1.EmuTable,EmuTable_tab.2.EmuTable,EmuTableUngrouped_tab.1.EmuTableUngrouped,EmuRef_tab.3.EmuRef.irn,EmuRef_tab.3.EmuRef.EmuRefOnly,EmuNestedTable_nesttab.1.EmuNestedTable,EmuNestedTable_nesttab.2.EmuNestedTable.1.2.EmuNestedTable",
            "1000000,Text,1,1.0,45 30 15 N,-130 10 5 W,1000000,Text,1970-01-01,Jan 1970,1970,9:00,12:00,15:00,Text,Text,Text,1000000,Text,,Text",
        ]


def test_write_xml_invalid_kind(rec, output_dir):
    with pytest.raises(ValueError, match="kind must be one of"):
        write_xml([rec], str(output_dir / "import.xml"), kind="invalid")


def test_group(rec, output_dir):
    rec.schema.validate_paths = False
    path = str(output_dir / "group.xml")
    write_group([rec], path, irn=1234567, name="Group")
    reader = EMuReader(path)
    for rec_ in reader:
        assert rec_ == {
            "GroupType": "Static",
            "Module": "emain",
            "Keys_tab": ["1000000"],
            "irn": "1234567",
            "GroupName": "Group",
        }
    rec.schema.validate_paths = True


def test_group_no_metadata(rec, output_dir):
    rec.schema.validate_paths = False
    with pytest.raises(ValueError, match="Must specify at least one of irn or name"):
        write_group([rec], str(output_dir / "group.xml"))
    rec.schema.validate_paths = True


def test_col_not_list(rec):
    with pytest.raises(TypeError, match="Columns must be lists"):
        rec["EmuTable_tab"] = ""


def test_rec_not_dict(rec):
    with pytest.raises(TypeError, match="References must be dicts"):
        rec["EmuRef"] = []


@pytest.mark.skip(reason="schema is no longer lazy loaded")
def test_rec_lazy_load_schema():
    EMuRecord.schema = None
    EMuRecord(module="ecatalogue")
    assert EMuRecord.schema


def test_rec_update():
    rec = EMuRecord(
        {
            "irn": 1000000,
            "EmuTable_tab(+)": ["Text", "Text", "Text"],
            "EmuRef_tab(+)": [
                {"irn": 1000000},
                {"irn": 1000000},
                {"irn": 1000001},
            ],
        },
        module="emain",
    )
    grid = rec.grid("EmuTable_tab")
    xml = etree.tostring(rec.to_xml()).decode("utf-8")
    assert xml.count('row="+"') == len(grid) * len(grid.columns)
    for row in grid:
        assert xml.count(f'group="{row.row_id()}"') == len(grid.columns)


def test_rec_get(rec):
    assert rec.get("EmuInvalid") is None


def test_rec_getitem_path(rec):
    assert rec["EmuRef.EmuRefOnly"] == "Text"


def test_rec_setdefault(rec):
    irn = rec["irn"]
    rec.setdefault("irn", 0)
    assert rec["irn"] == irn

    del rec["irn"]
    rec.setdefault("irn", 0)
    assert rec["irn"] == 0


def test_rec_irn_from_int(rec):
    irn = 1000000
    rec["EmuRef"] = irn
    rec["EmuRef_tab"] = [irn, irn]
    assert rec["EmuRef"] == irn
    assert rec["EmuRef_tab"] == [irn, irn]


def test_rec_getitem_empty_path(rec):
    with pytest.raises(KeyError, match=r"Path not found but valid"):
        rec["EmuEmpty"]


def test_rec_getitem_invalid_path(rec):
    with pytest.raises(KeyError, match=r"Invalid path"):
        rec["EmuInvalid"]


def test_rec_getitem_no_schema(rec):
    rec.schema = None
    with pytest.raises(KeyError, match=r"Path not found: EmuInvalid"):
        rec["EmuInvalid"]


@pytest.mark.parametrize(
    "key,val,match",
    [
        ("EmuText", ["Text"], r"Sequence assigned to atomic field"),
        ("EmuRef", "Text", r"References must be dicts"),
        ("EmuRef_tab", "1234567", r"Columns must be lists"),
        ("EmuRef_tab", ["Text"], r"Could not coerce to Integer"),
        ("EmuTable_tab", "Text", r"Columns must be lists"),
        ("EmuTable_tab", [["Text"]], r"Too many levels in a table"),
        ("EmuNestedTable_nesttab", "Text", r"Columns must be lists"),
        ("EmuNestedTable_nesttab", ["Text"], r"Columns must be lists"),
        ("EmuNestedTable_nesttab", [[["Text"]]], r"Too many levels in a nested table"),
    ],
)
def test_rec_set_invalid_type(key, val, match):
    with pytest.raises(TypeError, match=match):
        EMuRecord(module="emain")[key] = val


@pytest.mark.parametrize(
    "val,expected",
    [(1.0, 1), ("10,000", 10000), (-9999.0, -9999)],
)
def test_rec_set_integer(rec, val, expected):
    rec["EmuInteger"] = val
    assert rec["EmuInteger"] == expected


def test_rec_set_invalid_field():
    with pytest.raises(KeyError, match=r"Path not found:"):
        EMuRecord(module="emain")["EmuInvalid"] = 1


def test_rec_no_module():
    with pytest.raises(ValueError, match=r"Must provide module when schema is used"):
        EMuRecord()


@pytest.mark.parametrize(
    "date_string,kind,year,month,day,formatted,min_val,max_val",
    [
        ("2022-02-25", "day", 2022, 2, 25, "2022-02-25", "2022-02-25", "2022-02-25"),
        ("Feb 2022", "month", 2022, 2, None, "Feb 2022", "2022-02-01", "2022-02-28"),
        ("2022", "year", 2022, None, None, "2022", "2022-01-01", "2022-12-31"),
    ],
)
def test_dtype_date(date_string, kind, year, month, day, formatted, min_val, max_val):
    val = EMuDate(date_string)
    assert val.kind == kind
    assert val.year == year
    assert val.month == month
    assert val.day == day
    assert str(val) == formatted
    assert val.min_value == date(*[int(n) for n in min_val.split("-")])
    assert val.max_value == date(*[int(n) for n in max_val.split("-")])


def test_dtype_date_parse():
    date_from_str = EMuDate("2022-02-25")
    date_from_tuple = EMuDate((2022, 2, 25))
    date_from_args = EMuDate(2022, 2, 25)
    assert date_from_str == date_from_tuple == date_from_args


@pytest.mark.parametrize(
    "date_string,expected",
    [
        ("2022-01-31", False),
        ("2022-02-01", False),
        ("2022-02-15", False),
        ("2022-02-28", False),
        ("2022-03-01", False),
        ("Jan 2022", False),
        ("Feb 2022", True),
        ("Mar 2022", False),
        ("2021", False),
        ("2022", False),
        ("2023", False),
    ],
)
def test_dtype_date_eq(date_string, expected):
    assert (EMuDate("Feb 2022") == date_string) == expected


@pytest.mark.parametrize(
    "date_string,expected",
    [
        ("2022-01-31", True),
        ("2022-02-01", True),
        ("2022-02-15", True),
        ("2022-02-28", True),
        ("2022-03-01", True),
        ("Jan 2022", True),
        ("Feb 2022", False),
        ("Mar 2022", True),
        ("2021", True),
        ("2022", True),
        ("2023", True),
    ],
)
def test_dtype_date_ne(date_string, expected):
    assert (EMuDate("Feb 2022") != date_string) == expected


@pytest.mark.parametrize(
    "date_string,expected",
    [
        ("2022-01-31", False),
        ("2022-02-01", False),
        ("2022-02-15", False),
        ("2022-02-28", False),
        ("2022-03-01", True),
        ("Jan 2022", False),
        ("Feb 2022", False),
        ("Mar 2022", True),
        ("2021", False),
        ("2022", False),
        ("2023", True),
    ],
)
def test_dtype_lt(date_string, expected):
    assert (EMuDate("Feb 2022") < date_string) == expected


@pytest.mark.parametrize(
    "date_string,expected",
    [
        ("2022-01-31", False),
        ("2022-02-01", True),
        ("2022-02-15", True),
        ("2022-02-28", True),
        ("2022-03-01", True),
        ("Jan 2022", False),
        ("Feb 2022", True),
        ("Mar 2022", True),
        ("2021", False),
        ("2022", True),
        ("2023", True),
    ],
)
def test_dtype_le(date_string, expected):
    assert (EMuDate("Feb 2022") <= date_string) == expected


@pytest.mark.parametrize(
    "date_string,expected",
    [
        ("2022-01-31", True),
        ("2022-02-01", False),
        ("2022-02-15", False),
        ("2022-02-28", False),
        ("2022-03-01", False),
        ("Jan 2022", True),
        ("Feb 2022", False),
        ("Mar 2022", False),
        ("2021", True),
        ("2022", False),
        ("2023", False),
    ],
)
def test_dtype_gt(date_string, expected):
    assert (EMuDate("Feb 2022") > date_string) == expected


@pytest.mark.parametrize(
    "date_string,expected",
    [
        ("2022-01-31", True),
        ("2022-02-01", True),
        ("2022-02-15", True),
        ("2022-02-28", True),
        ("2022-03-01", False),
        ("Jan 2022", True),
        ("Feb 2022", True),
        ("Mar 2022", False),
        ("2021", True),
        ("2022", True),
        ("2023", False),
    ],
)
def test_dtype_ge(date_string, expected):
    assert (EMuDate("Feb 2022") >= date_string) == expected


def test_dtype_invalid_comp():
    assert (EMuFloat(0) == None) == False
    assert (EMuFloat(0) != None) == True
    with pytest.raises(TypeError, match=r"'<' not supported"):
        assert (EMuFloat(0) < None) == False
    with pytest.raises(TypeError, match=r"'<=' not supported"):
        assert (EMuFloat(0) <= None) == False
    with pytest.raises(TypeError, match=r"'>' not supported"):
        assert (EMuFloat(0) > None) == False
    with pytest.raises(TypeError, match=r"'>=' not supported"):
        assert (EMuFloat(0) >= None) == False


@pytest.mark.parametrize(
    "date_string,expected",
    [
        ("2022-01-31", False),
        ("2022-02-01", True),
        ("2022-02-15", True),
        ("2022-02-28", True),
        ("2022-03-01", False),
        ("Jan 2022", False),
        ("Feb 2022", True),
        ("Mar 2022", False),
        ("2021", False),
        ("2022", False),
        ("2023", False),
        ((2022, 2, 1), True),
    ],
)
def test_dtype_contains(date_string, expected):
    assert (date_string in EMuDate("Feb 2022")) == expected


@pytest.mark.parametrize(
    "date_string,exp_str,exp_emu",
    [
        ("99999-01-01", "99999-01-01", "99999-01-01"),
        ("99999-01-", "Jan 99999", "Jan 99999"),
        ("Jan 99999", "Jan 99999", "Jan 99999"),
        ("99999", "99999", "99999"),
        (99999, "99999", "99999"),
        ("-99999-01-01", "-99999-01-01", "99999-01-01 BC"),
        ("-99999-01-", "Jan -99999", "Jan 99999 BC"),
        ("Jan -99999", "Jan -99999", "Jan 99999 BC"),
        ("-99999", "-99999", "99999 BC"),
        (-99999, "-99999", "99999 BC"),
        ("99999-01-01 BC", "-99999-01-01", "99999-01-01 BC"),
        ("99999-01- b.c.", "Jan -99999", "Jan 99999 BC"),
        ("Jan 99999 bce", "Jan -99999", "Jan 99999 BC"),
        ("99999 B.C.", "-99999", "99999 BC"),
        (99, "0099", "0099 AD"),
        (9, "0009", "0009 AD"),
        (0, "0000", "0000 AD"),
        (-9, "-0009", "0009 BC"),
        (-99, "-0099", "0099 BC"),
    ],
)
def test_dtype_date_out_of_range(date_string, exp_str, exp_emu):
    emu_date = EMuDate(date_string)
    assert str(emu_date) == exp_str
    assert emu_date.emu_str() == exp_emu


def test_dtype_date_parse_same_class():
    emu_date = EMuDate("1970-01-01")
    assert emu_date == EMuDate(emu_date)


def test_dtype_date_operations():
    val = EMuDate("Jan 1970")
    assert val + timedelta(days=1) == (EMuDate("1970-01-02"), EMuDate("1970-02-01"))
    assert val - timedelta(days=1) == (EMuDate("1969-12-31"), EMuDate("1970-01-30"))
    assert EMuDate("1970-01-02") - EMuDate("1970-01-01") == timedelta(days=1)


def test_dtype_date_parse_failed():
    with pytest.raises(ValueError, match="Could not parse date:"):
        EMuDate("Jan 1, 1970")


def test_dtype_date_invalid_directive():
    date = EMuDate("1970-01-")
    with pytest.raises(ValueError, match=r"Invalid directives for \(1970, 1, None\)"):
        date.strftime("%Y-%m-%d")


@pytest.mark.skip(reason="attribute modification no longer permitted")
@pytest.mark.parametrize("attr", ["min_value", "max_value"])
def test_dtype_date_bad_kind(attr):
    date = EMuDate("1970-01-01")
    date.kind = None
    with pytest.raises(ValueError, match="Invalid kind:"):
        getattr(date, attr)


def test_dtype_date_to_datetime():
    date = EMuDate("1970-01-01")
    time = EMuTime("15:00")
    assert date.to_datetime(time) == datetime(1970, 1, 1, 15, 0)
    assert time.to_datetime(date) == datetime(1970, 1, 1, 15, 0)


@pytest.mark.parametrize(
    "date_tuple,expected",
    [
        ((1970, 1, 1), "1970-01-01"),
        ((1970, 1, None), "Jan 1970"),
        ((1970, None, None), "1970"),
    ],
)
def test_dtype_date_tuples(date_tuple, expected):
    assert str(EMuDate(date_tuple)) == expected
    assert str(EMuDate(ExtendedDate(*date_tuple))) == expected


@pytest.mark.parametrize(
    "date_tuple,match",
    [
        ((10000, 13, 1), r"Month out of range"),
        ((10000, 1, 32), r"Day out of range"),
    ],
)
def test_dtype_date_invalid(date_tuple, match):
    with pytest.raises(ValueError, match=match):
        EMuDate(date_tuple)


@pytest.mark.skip(reason="attribute modification no longer permitted")
def test_dtype_date_setters():
    date = EMuDate("1970-01-01")
    date.year = 1971
    date.month = 2
    date.day = 2
    assert str(date) == "1971-02-02"


@pytest.mark.parametrize(
    "time_string",
    [
        "1500",
        "15:00",
        "3:00 PM",
        "0300 PM",
        "15:00 UTC-0700",
        "15:00 -0700",
        "3:00 PM UTC-0700",
        "3:00 PM -0700",
        time(hour=15, minute=0),
        EMuTime("1500"),
    ],
)
def test_dtype_time(time_string):
    time = EMuTime(time_string)
    assert time.hour == 15
    assert time.minute == 0


@pytest.mark.parametrize(
    "val,fmt,expected",
    [
        ("1.10", None, "1.10"),
        (1.10, "{:.2f}", "1.10"),
        (EMuFloat("1.10"), None, "1.10"),
        ("1.", None, "1"),
    ],
)
def test_dtype_float_str(val, fmt, expected):
    assert str(EMuFloat(val, fmt)) == expected


@pytest.mark.parametrize(
    "cl,val,expected",
    [
        (EMuFloat, "1.1", "EMuFloat('1.1')"),
        (EMuFloat, "1.10", "EMuFloat('1.10')"),
        (EMuFloat, "1.100", "EMuFloat('1.100')"),
        (EMuLatitude, "45°30'15''N", "EMuLatitude('45 30 15 N')"),
        (EMuLongitude, "45.5042", "EMuLongitude('45.5042')"),
        (EMuLongitude, "-45.5042", "EMuLongitude('-45.5042')"),
    ],
)
def test_dtype_coord_repr(cl, val, expected):
    assert repr(cl(val)) == expected


def test_dtype_float_format():
    val = EMuFloat("1.10")
    assert "{}".format(val) == "1.10"
    assert "{:.1f}".format(val) == "1.1"
    assert "{:.3f}".format(val) == "1.100"

    assert f"{val}" == "1.10"
    assert f"{val:.1f}" == "1.1"
    assert f"{val:.3f}" == "1.100"


def test_dtype_float_operations():
    val = EMuFloat("0.1200")

    assert val + 1 == pytest.approx(1.12)
    assert val - 0.12 == pytest.approx(0)
    assert val * 2 == pytest.approx(0.24)
    assert val / 2 == pytest.approx(0.06)
    assert val // 2 == pytest.approx(0)
    assert val % 0.12 == pytest.approx(0)
    assert divmod(val, 0.12) == (1, 0)
    assert val**2 == pytest.approx(0.0144)

    assert val == 0.12
    assert val != 0.121
    assert val < 1
    assert val <= 0.12
    assert val > 0
    assert val >= 0.12

    assert str(val) == "0.1200"

    val %= 1
    assert val == pytest.approx(0.12)
    val += 1
    assert val == pytest.approx(1.12)
    val -= 1
    assert val == pytest.approx(0.12)
    val *= 2
    assert val == pytest.approx(0.24)
    val /= 2
    assert val == pytest.approx(0.12)
    val **= 2
    assert val == pytest.approx(0.0144)
    val //= 2
    assert val == pytest.approx(0)


def test_dtype_float_sigfigs():
    val = EMuFloat("0.1200")
    assert str(val + EMuFloat("1.00")) == "1.1200"
    assert str(val * EMuFloat("2.00")) == "0.24"


def test_dtype_float_type_conversions():
    val = EMuFloat("0.1200")
    assert str(val) == str(EMuFloat(0.1200, "{:.4f}"))
    assert int(val) == 0
    assert float(val) == 0.12


def test_dtype_float_values():
    val = EMuFloat("0.1200")
    assert val.value == 0.12
    assert val.min_value == 0.12
    assert val.max_value == 0.12
    assert val.comp == 0.12
    assert val.min_comp == 0.12
    assert val.max_comp == 0.12


def test_dtype_float_mutability():
    val = EMuFloat("1.0")
    val_id = id(val)
    val += 1
    assert val_id != id(val)


def test_dtype_setattr():
    with pytest.raises(AttributeError, match="Cannot modify existing attribute"):
        EMuFloat("0.1200").value = 0.1


def test_dtype_delattr():
    with pytest.raises(AttributeError, match="Cannot delete attribute"):
        del EMuFloat("0.1200").value


@pytest.mark.skip(reason="no longer the expected behavior")
def test_dtype_float_no_format():
    with pytest.raises(ValueError, match="Must provide fmt when passing a float"):
        EMuFloat(0.12)


def test_dtype_float_contain_no_range():
    with pytest.raises(ValueError, match="EMuFloat is not a range"):
        1 in EMuFloat("0.12")


@pytest.mark.parametrize(
    "val,fmt",
    [
        ("45°30'15''N", None),
        ("45 30 15 North", None),
        ("N 45 30 15", None),
        ("45 30.25 N", None),
        ("45.5042", None),
        (45.5042, "{:.4f}"),
    ],
)
def test_dtype_latitude(val, fmt):
    lat = EMuLatitude(val, fmt=fmt)
    assert float(lat) == pytest.approx(45.5042)
    assert int(lat) == 45
    assert lat.to_dec() == "45.5042"
    assert lat.to_dms() == "45 30 15 N"


@pytest.mark.parametrize(
    "val,fmt",
    [
        ("45°30'15''W", None),
        ("45 30 15 West", None),
        ("W 45 30 15", None),
        ("45 30.25 W", None),
        ("-45.5042", None),
        (-45.5042, "{:.4f}"),
    ],
)
def test_dtype_longitude(val, fmt):
    lng = EMuLongitude(val, fmt=fmt)
    assert float(lng) == pytest.approx(-45.5042)
    assert int(lng) == -45
    assert lng.to_dec() == "-45.5042"
    assert lng.to_dms() == "45 30 15 W"


@pytest.mark.parametrize(
    "val,unc_m,expected_dms,expected_dec",
    [
        ("45 30 15 N", 10, "45 30 15 N", "45.5042"),
        ("45 30 15 N", 20, "45 30 15 N", "45.5042"),
        ("45 30 15 N", 50, "45 30.3 N", "45.504"),
        ("45 30 15 N", 90, "45 30.3 N", "45.504"),
        ("45 30 15 N", 100, "45 30.3 N", "45.504"),
        ("45 30 15 N", 200, "45 30.3 N", "45.504"),
        ("45 30 15 N", 500, "45 30 N", "45.50"),
        ("45 30 15 N", 900, "45 30 N", "45.50"),
        ("45 30 15 N", 1000, "45 30 N", "45.50"),
    ],
)
def test_dtype_coord_rounding(val, unc_m, expected_dms, expected_dec):
    lat = EMuLatitude(val)
    assert lat.to_dms(unc_m) == expected_dms
    assert lat.to_dec(unc_m) == expected_dec


@pytest.mark.parametrize("coord_class", [EMuLatitude, EMuLongitude])
def test_dtype_coord_invalid(coord_class):
    with pytest.raises(ValueError, match=r"Invalid coordinate"):
        coord_class("45 30 15 7.5 N")


@pytest.mark.parametrize(
    "coord_class,val", [(EMuLatitude, "90.1"), (EMuLongitude, "-180.1")]
)
def test_dtype_coord_out_of_bounds(coord_class, val):
    with pytest.raises(ValueError, match=r"Coordinate out of bounds"):
        coord_class(val)


def test_dtype_coord_to_dms_too_precise():
    with pytest.raises(ValueError, match=r"unc_m cannot be smaller"):
        EMuLatitude("45 30 15 N").to_dms(1)


def test_dtype_coord_to_dec_too_precise():
    with pytest.raises(ValueError, match=r"unc_m cannot be smaller"):
        EMuLongitude("45 30 15 E").to_dec(1)


def test_dtype_coord_unsigned():
    with pytest.raises(ValueError, match=r"Could not parse as EMuLatitude"):
        EMuLatitude("45 30 15")


def test_dtype_coord_invalid_minutes():
    with pytest.raises(ValueError, match=r"Invalid minutes"):
        EMuLongitude("45 90 15 E")


def test_dtype_coord_invalid_seconds():
    with pytest.raises(ValueError, match=r"Invalid seconds"):
        EMuLongitude("45 30 75 E")


@pytest.mark.parametrize(
    "field,expected",
    [
        ("AtomField", False),
        ("AtomFieldRef", False),
        ("TableField0", True),
        ("TableField_nesttab", True),
        ("TableField_nesttab_inner", True),
        ("TableFieldRef_tab", True),
        ("TableField_tab(+)", True),
        ("TableField_tab(-)", True),
        ("TableField_tab(=)", True),
        ("TableField_tab(1+)", True),
        ("TableField_tab(12-)", True),
        ("TableField_tab(123=)", True),
    ],
)
def test_is_tab(field, expected):
    assert is_tab(field) == expected


@pytest.mark.parametrize(
    "field,expected",
    [
        ("AtomField", False),
        ("AtomFieldRef", False),
        ("TableField0", False),
        ("TableField_nesttab", True),
        ("TableField_nesttab_inner", False),
        ("TableFieldRef_tab", False),
        ("TableField_tab(+)", False),
        ("TableField_tab(-)", False),
        ("TableField_tab(=)", False),
        ("TableField_tab(1+)", False),
        ("TableField_tab(12-)", False),
        ("TableField_tab(123=)", False),
    ],
)
def test_is_nesttab(field, expected):
    assert is_nesttab(field) == expected


@pytest.mark.parametrize(
    "field,expected",
    [
        ("AtomField", False),
        ("AtomFieldRef", False),
        ("TableField0", False),
        ("TableField_nesttab", False),
        ("TableField_nesttab_inner", True),
        ("TableFieldRef_tab", False),
        ("TableField_tab(+)", False),
        ("TableField_tab(-)", False),
        ("TableField_tab(=)", False),
        ("TableField_tab(1+)", False),
        ("TableField_tab(12-)", False),
        ("TableField_tab(123=)", False),
    ],
)
def test_is_nesttab_inner(field, expected):
    assert is_nesttab_inner(field) == expected


@pytest.mark.parametrize(
    "field,expected",
    [
        ("AtomField", False),
        ("AtomFieldRef", True),
        ("TableField0", False),
        ("TableField_nesttab", False),
        ("TableField_nesttab_inner", False),
        ("TableFieldRef_tab", True),
        ("TableFieldRef_tab(+)", True),
        ("TableFieldRef_tab(-)", True),
        ("TableFieldRef_tab(=)", True),
        ("TableFieldRef_tab(1+)", True),
        ("TableFieldRef_tab(12-)", True),
        ("TableFieldRef_tab(123=)", True),
    ],
)
def test_is_ref(field, expected):
    assert is_ref(field) == expected


@pytest.mark.parametrize(
    "field,expected",
    [
        ("AtomField", False),
        ("AtomFieldRef", False),
        ("TableField0", False),
        ("TableField_nesttab", False),
        ("TableField_nesttab_inner", False),
        ("TableFieldRef_tab", True),
        ("TableField_tab", False),
        ("TableFieldRef_tab(+)", True),
        ("TableFieldRef_tab(-)", True),
        ("TableFieldRef_tab(=)", True),
        ("TableFieldRef_tab(1+)", True),
        ("TableFieldRef_tab(12-)", True),
        ("TableFieldRef_tab(123=)", True),
    ],
)
def test_is_ref_tab(field, expected):
    assert is_ref_tab(field) == expected


@pytest.mark.parametrize(
    "field,expected",
    [
        ("AtomField", "AtomField"),
        ("AtomFieldRef", "AtomFieldRef"),
        ("TableField0", "TableField"),
        ("TableField_nesttab", "TableField"),
        ("TableField_nesttab_inner", "TableField"),
        ("TableFieldRef_tab", "TableFieldRef"),
        ("TableField_tab", "TableField"),
        ("TableField_tab(+)", "TableField"),
        ("TableField_tab(-)", "TableField"),
        ("TableField_tab(=)", "TableField"),
        ("TableField_tab(1+)", "TableField"),
        ("TableField_tab(12-)", "TableField"),
        ("TableField_tab(123=)", "TableField"),
    ],
)
def test_strip_tab(field, expected):
    assert strip_tab(field) == expected


@pytest.mark.parametrize(
    "field",
    [
        "TableField0",
        "TableField_nesttab",
        "TableField_nesttab_inner",
        "TableFieldRef_tab",
        "TableField_tab",
    ],
)
def test_mods(field):
    for mod in ["+", "-", "="]:
        for num in range(5):
            mod = f"{2 ** num if num else ''}{mod}"
            field_with_mod = f"{field}({mod})"
            assert has_mod(field_with_mod)
            assert strip_mod(field_with_mod) == field
            assert get_mod(field_with_mod) == mod


def test_mod_on_atom():
    with pytest.raises(ValueError, match=r"Update modifier found on an atomic"):
        has_mod("AtomField(+)")


def test_mod_invalid():
    with pytest.raises(ValueError, match=r"Invalid modifier"):
        get_mod("AtomField(*)")


def test_pickle(rec):
    assert rec == pickle.loads(pickle.dumps(rec))


def test_json(rec):
    assert rec == EMuRecord(rec.json(), module=rec.module)


def test_encoder(rec):
    dct_from_json = json.loads(json.dumps(rec, cls=EMuEncoder))
    assert rec == EMuRecord(dct_from_json, module=rec.module)


@pytest.mark.parametrize("kind", ["class", "copy"])
def test_mutability(kind, rec):
    if kind == "class":
        rec_copy = EMuRecord(rec, module=rec.module)
    else:
        rec_copy = rec.copy()

    for key, val in rec.items():
        assert val == rec_copy[key]

    rec["EmuText"] += "s"
    assert rec["EmuText"] == "Texts"
    assert rec_copy["EmuText"] == "Text"

    rec["EmuFloat"] += 1
    assert rec["EmuFloat"] == 2
    assert rec_copy["EmuFloat"] == 1

    rec["EmuRef"]["irn"] = 1000001
    assert rec_copy["EmuRef"]["irn"] == 1000000

    del rec["EmuDate0"][0]
    assert rec["EmuDate0"][0] == EMuDate("Jan 1970")
    assert rec_copy["EmuDate0"][0] == EMuDate("1970-01-01")

    rec["EmuNestedTable_nesttab"][0].append("Text")
    rec["EmuNestedTable_nesttab"][0][0] == "Text"
    rec["EmuNestedTable_nesttab"][0] == []
